=head1 NAME

 TransDB::Root

=head1 DESCRIPTION

 A basic object from which all others can inherit. Provides an
 interface and general methods such as object creation, an
 AUTOLOAD function for non-specialised get/set, and warnings
 with stack backtrace.

=head1 EXAMPLE USAGE

 package Child;

 use strict;
 use vars qw(@ISA);

 use TransDB::Root;

 @ISA = qw(TransDB::Root);

 my %cols = (); # needed for storage of attribute names
 Child->attributes(qw(id name description));

 sub _initialise {
     my($self, $id, $name, $description) = @_;

     $self->id($id);
     $self->name($name);
     $self->description($description);

     return $self;
 }

 1;

=head1 AUTHOR - Matthew Betts

 mailto:matthew.betts@bioquant.uni-heidelberg.de

=head1 APPENDIX

 The rest of the documentation details each of the object methods.
 Internal methods normally begin with an underscore.

=cut

package TransDB::Root;

use strict;
use vars qw(@ISA $AUTOLOAD $VERSION);
use Carp;
use Class::ISA;

$VERSION = '1.6b';

=head2 attributes

  Usage   : called within a derived class (see example above)
  Function: sets up attributes available to a derived class
  Returns : a list of names of all attributes, including ones from ancestor classes
  Args    : names of the attributes

  Default get/set methods are automatically provided with the same
  names as the attributes (via AUTOLOAD).

  Done this way so that named attributes can be used with perl lists
  rather than hashes. The names of the attributes are stored once for
  each class rather than once for each object. This can save quite a
  lot of memory, especially for attributes with long names. 10000
  anonymous hashes with 10 named variables each takes about 1.4 x the
  memory of 10000 anonymous lists with ten elements (variable names
  five characters in length).

=cut

my %cols = ();

sub attributes {
    my($class, @keys) = @_;

    my $all_keys;
    my $super;
    my $i;
    my $key;

    defined($cols{$class}) or ($cols{$class} = {});

    #@{$cols{$class}}{@keys} = (0..$#keys);

    # the column index of a key should not be the same as that of
    # the key of an ancestor class, unless it has the same name

    # make sure the keys of ancestor classes are processed first
    $all_keys = {};
    foreach $super (reverse(Class::ISA::super_path($class))) {
	foreach $key (keys %{$cols{$super}}) {
	    $all_keys->{$key} = $cols{$super}->{$key};
	}
    }

    # assign column indexes to the new keys, processing them in the given order
    $i = scalar keys %{$all_keys};
    foreach $key (@keys) {
	if(!defined($all_keys->{$key})) {
	    $cols{$class}->{$key} = $i;
	    $all_keys->{$key} = $i;
	    ++$i;
	}
    }

    #print join("\t", $class, sort {$all_keys->{$a} <=> $all_keys->{$b}} keys(%{$all_keys})), "\n";

    return(keys(%{$all_keys}));
}

=head2 _col

  Usage   :
  Function:
  Returns :
  Args    :

=cut

sub _col {
    my($class, $key) = @_;

    my $col;
    my $super;

    if(!defined($col = $cols{$class}->{$key})) {
	foreach $super (Class::ISA::super_path($class)) {
	    $col = $cols{$super}->{$key};
	    defined($col) and last;
	}
    }

    return $col;
}

=head2 new

  Usage   : depends on the child class
  Function: creates an object as a hash, bless the hash, and initialises the object
  Returns : the return value of _initialise, which should be the object if successful and undef otherwise
  Args    : depends on the child class

=cut

sub new {
    my($class, @args) = @_;

    my $self;

    $self = [];
    bless $self, $class;
    $self = $self->_initialise(@args);

    return $self;
}

=head2 _initialise

  Usage   : usually called only from within new()
  Function: simple initialiser for '-key' => 'value' pairs.
            Overload for more complex initialisation.
  Returns : the calling object
  Args    : key-value pairs

=cut

sub _initialise {
    my($self, %args) = @_;

    my $class;
    my $key;
    my %args2;
    my $key2;
    my $value;

    $class = ref $self;

    %args2 = ();
    foreach $key (keys %args) {
	($key2 = $key) =~ s/^-//;
	$args2{$key2} = $args{$key};
    }

    foreach $key (sort {$class->_col($a) <=> $class->_col($b)} keys %args2) { # processed in the order given by attributes()
	$value = $args2{$key};
	$key =~ s/^-//;
	$self->$key($value);
    }

    return $self;
}

=head2 AUTOLOAD

  Usage   : default when calling an unknown method
  Function: non-specialised get/set on object variables. Assumes that
            the object is a hash and that the variables are values in
            that hash with a particular key.
  Returns : the value of the variable
  Args    : the value of the variable (if setting it)

=cut

sub AUTOLOAD {
    my($self, $value) = @_;

    my $class;
    my $key;
    my $col;

    $AUTOLOAD =~ /(.*)::(.*)/;
    $class = $1;
    $key = $2;

    if(defined($col = $class->_col($key))) {
	if(defined($value)) {
	    $self->[$col] = $value;
	}
	else {
	    $value = $self->[$col];
	}
    }
    else {
	$self->warn("${class}::$key not implemented by AUTOLOAD");
	return undef;
    }

    return $value;
}

=head2 warn

  Usage   : $self->warn("arrrgghhhh!");
  Function: warnings with stack backtrace
  Returns : nothing
  Args    : warning message

=cut

sub warn {
    my($self, @messages) = @_;

    Carp::cluck('Warning: ', @messages);
}

=head2 DESTROY

  Usage   : called automatically when an object goes out of scope
  Function: nothing, here only to prevent AUTOLOAD from complaining.
            Can be overloaded by child classes if necessary (for
            example to remove temporary files).
  Returns : nothing
  Args    : none

=cut

sub DESTROY {
}

1;
