package TransDB::Utils;

use strict;
use vars qw(@ISA);

use TransDB::Root;
use File::Copy;
use Bio::SeqIO;

@ISA = qw(TransDB::Root);

sub job_id {
    my($dn_tmp) = @_;

    my $dn;
    my $id0;
    my $id1;
    my $i;
    my $attempt;
    my $success;

    if(defined($ENV{'PBS_ARRAY_ID'})) {
	($id0 = $ENV{'PBS_ARRAY_ID'}) =~ s/\[.*//;
	defined($ENV{'PBS_ARRAY_INDEX'}) and ($id0 = join('_', $id0, $ENV{'PBS_ARRAY_INDEX'}));
    }
    elsif(defined($ENV{'PBS_JOBID'})) {
	($id0 = $ENV{'PBS_JOBID'}) =~ s/\A(\d+).*/$1/;
	defined($ENV{'PBS_JOBNAME'}) and ($id0 = join('_', $id0, $ENV{'PBS_JOBNAME'}));
    }
    else {
	$id0 = time;
    }
    $id1 = $id0;

    # make sure the job id is not already in use
    $dn = join '', $dn_tmp, $id1, '/';
    $i = 0;
    while(-e $dn) {
	$id1 = join '_', $id0, ++$i;
	$dn = join '', $dn_tmp, $id1, '/';
    }

    # very occasionally the above check will pass for two jobs with
    # the same id, presumably if they're checking at the same time.
    # This will mean one of them will be unable to make the directory.
    $attempt = 0;
    $success = mkdir($dn);
    while(!$success and ($attempt < 10)) {
	sleep 5;
	++$attempt;
	$id1 = join '_', $id0, ++$i;
	$dn = join '', $dn_tmp, $id1, '/';
	$success = mkdir($dn);
    }

    if(!$success) {
	TransDB::Root::warn("Error: cannot mkdir $dn: $!");
	return undef;
    }

    # FIXME - if several jobs are running on different machines but end up copying
    # results back to the same directory, and these jobs happen to have the same
    # job id, then some of the results will be overwritten. This is avoided for
    # the cluster, but might need attention if running on other systems.

    return($id1, $dn);
}

sub openfile {
    my($fn) = @_;

    my $fh;

    # FIXME - only opens files for reading at the moment

    if(($fn =~ /\.gz\Z/) or ($fn =~ /\.Z\Z/)) {
	if(!open($fh, "zcat $fn |")) {
	    TransDB::Root::warn("cannot open pipe from 'zcat $fn'.");
	    $fh = undef;
	}
    }
    elsif($fn =~ /\.bz2\Z/) {
	if(!open($fh, "bzip2 -d -c $fn |")) {
	    TransDB::Root::warn("cannot open pipe from 'bzip2 -d -c $fn'.");
	    $fh = undef;
	}
    }
    else {
	if(!open($fh, $fn)) {
	    TransDB::Root::warn("cannot open '$fn' file for reading.");
	    $fh = undef;
	}
    }

    return $fh;
}

sub copy_dom2dn {
    my($fn, $dn) = @_;

    my $fh_in;
    my $fh_out;
    my $fn_out;
    my $fn_pdb;
    my $fn_pdb1;
    my $id;
    my $dom;

    ($fn_out = $fn) =~ s/\A.*\///;
    $fn_out = $dn . $fn_out;

    if(!open($fh_out, ">$fn_out")) {
	TransDB::Root::warn("cannot open '$fn_out' file for writing.");
	return undef;
    }

    if(!open($fh_in, $fn)) {
	TransDB::Root::warn("Error: copy_dom2dn: cannot open '$fn' file for reading.");
	return undef;
    }
    while(<$fh_in>) {
	if(/\A(\S+)\s+(\S+)\s+(\{.*)\Z/) {
	    ($fn_pdb, $id, $dom) = ($1, $2, $3);
	    $fn_pdb1 = join '', $dn, $id, '.pdb';
	    ($fn_pdb =~ /\.gz\Z/) and ($fn_pdb1 .= '.gz');
	    if(!(-e $fn_pdb1)) {
		if(!copy($fn_pdb, $fn_pdb1)) {
		    TransDB::Root::warn("Error: cannot copy $fn_pdb to $fn_pdb1.");
		    return undef;
		}
	    }
	    print $fh_out "$fn_pdb1 $id $dom\n";
	}
	else {
	    print $fh_out $_;
	}
    }
    close($fh_in);
    close($fh_out);

    return $fn_out;
}

sub copy2dn {
    my($fn, $dn, $force) = @_;

    my $fn1;

    ($fn1 = $fn) =~ s/\A.*\///;
    $fn1 = $dn . $fn1;
    if(!(-e $fn1) or $force) {
	if(!copy($fn, $fn1)) {
	    TransDB::Root::warn("Error: cannot copy $fn to $fn1: $!");
	    return undef;
	}
    }

    return $fn1;
}

sub cleandir {
    my($dn) = @_;

    my $dh;
    my $fn;

    if(!opendir($dh, $dn)) {
	TransDB::Root::warn("Error: cannot open '$dn' for reading.");
	return 0;
    }

    while($fn = readdir $dh) {
	if($fn !~ /\A\.{1,2}\Z/) {
	    unlink("$dn/$fn") or warn("Error: cannot unlink $dn/$fn.");
	}
    }
    closedir($dh);
    if(!rmdir($dn)) {
	TransDB::Root::warn("Error: cannot rmdir $dn.");
	return 0;
    }

    return 1;
}

sub pdbseq {
    my($fn_dom, $out_fasta, $ids) = @_;

    my $cmd;
    my $fh_pdbseq;
    my @errs;
    my $seqs;
    my $out;
    my $seq;
    my $id;
    my $id2;
    my $str;

    $cmd = "pdbseq -f $fn_dom -min 0 -tl 0";

    if(!open($fh_pdbseq, "$cmd |")) {
	TransDB::Root::warn("Error: pdbseq: '$cmd' failed.");
	return undef;
    }

    # pdbseq returns 0 even when there are errors.
    @errs = ();
    $ids = {};
    $seqs = {};
    while(<$fh_pdbseq>) {
	if(/^error:\s*(.*)/) {
	    push @errs, "Error: pdbseq: $1";
	}
	elsif(/^>(\S+)/) {
	    $id2 = $1;

	    if(defined($id)) {
		if($str =~ /\A\s*\Z/) {
		    push @errs, "Error: pdbseq: zero-length sequence for '" . $seq->id . "'";
		}
		else {
		    $seq->seq($str);
		    defined($out_fasta) and $out_fasta->write_seq($seq);
		    $seqs->{$id} = $str;
		}
	    }

	    $id = $id2;
	    $ids->{$id}++;
	    $str = '';

	    $seq = Bio::Seq->new('-id' => $id, '-desc' => '');
	}
	else {
	    chomp;
	    $str .= $_;
	}
    }
    close($fh_pdbseq);

    if(defined($id)) {
	if($str =~ /\A\s*\Z/) {
	    push @errs, "Error: pdbseq: zero-length sequence for '" . $seq->id . "'";
	}
	else {
	    $seq->seq($str);
	    defined($out_fasta) and $out_fasta->write_seq($seq);
	    $seqs->{$id} = $str;
	}
    }

    foreach $str (@errs) {
	TransDB::Root::warn($str);
	return undef;
    }
    foreach $id (keys %{$ids}) {
	($ids->{$id} > 0) or warn("Error: pdbseq: no sequence for '$id'");
    }

    return $seqs;
}

sub cofm {
    my($fn_dom, $fn_cofm) = @_;

    my $fn_pdb;
    my $id;
    my $dom;
    my $fh_dom;
    my $cmd;
    my $fh;
    my $cofms;
    my $cofm;
    my $stat;
    my $fn_dom1;
    my $fh_dom1;
    my $fn_err;
    my $fh_err;
    my $x;
    my $y;
    my $z;

    if(!open($fh_dom, $fn_dom)) {
	TransDB::Root::warn("Error: cofm cannot open '$fn_dom' file for reading.");
	return undef;
    }
    $cofms = {};
    $fn_err = $fn_cofm . '.err';
    while(<$fh_dom>) {
	if(/\A(\S+)\s+(\S+)\s+\{\s+(.*?)\s*\}{0,1}\Z/) {
	    ($fn_pdb, $id, $dom) = ($1, $2, $3);

	    $fn_dom1 = $fn_dom . '.1';
	    if(!open($fh_dom1, ">$fn_dom1")) {
		TransDB::Root::warn("Error: cofm: cannot open '$fn_dom1' file for writing.");
		next;
	    }
	    printf $fh_dom1 "%s %s { %s }\n", $fn_pdb, $id, $dom;
	    close($fh_dom1);

	    $cmd = "cofm -f $fn_dom1 1> $fn_cofm 2> $fn_err";
	    $stat = system($cmd);

	    # cofm doesn't exit with 0 even when there are no errors...
#	    $stat >>= 8;
#	    if($stat != 0) {
#		TransDB::Root::warn("Error: cofm failed: ", $!);
#		next;
#	    }

	    # ... so check if the error file has anything in it
	    if(-s $fn_err) {
		TransDB::Root::warn("Error: cofm failed on $fn_dom1.");
		next;
	    }

	    if(!open($fh, $fn_cofm)) {
		TransDB::Root::warn("Error: TransDB::Utils::cofm: cannot open '$fn_cofm' file for reading."); # FIXME - use TransDB::Root::warn
		next;
	    }
	    while(<$fh>) {
		if(/^REMARK Domain\s+\S+ Id\s+(\S+) N =\s+\d+ Rg =\s+(\S+) Rmax =\s+(\S+) Ro\s+=\s+(\S+)\s+(\S+)\s+(\S+)/) {
		    $cofm = {
			     'id'     => $1,
			     'Rg'     => $2,
			     'Rmax'   => $3,
			     'Cx'     => $4,
			     'Cy'     => $5,
			     'Cz'     => $6,
			     'atoms'  => '',
			     'coords' => [],
			    };
		    $cofms->{$cofm->{'id'}} = $cofm;
		}
		elsif(/^ATOM.{26}(.{8})(.{8})(.{8})/) {
		    ($x, $y, $z) = ($1, $2, $3);

		    $x =~ s/\s*//g;
		    $y =~ s/\s*//g;
		    $z =~ s/\s*//g;

		    $cofm->{'atoms'} .= $_;
		    push @{$cofm->{'coords'}}, $x, $y, $z;
		}
	    }
	    close($fh);
	}
    }

    return $cofms;
}

sub guess_blast_format {
    my($fn) = @_;

    my $fh;
    my $m;
    my @F;

    if(!open($fh, $fn)) {
	TransDB::Root::warn("Error: guess_blast_format: cannot open '$fn' file for reading.");
	return undef;
    }

    $m = undef;
    while(<$fh>) {
	(/^#/ or /\A\s*\Z/) and next;

	if(/^BLAST/) {
	    $m = 0;
	    last;
	}
	else {
	    @F = split /\t/;
	    if(@F == 12) {
		$m = 9;
		last;
	    }
	}
    }

    return $m;
}

sub overlap {
    my($start0, $end0, $start1, $end1) = @_;

    my $overlap;
    my $dl; # left overhang (-1 if no overlap)
    my $dr; # right overhang (-1 if no overlap)
    my $type;

    if($start0 <= $start1) {
	if($end0 >= $start1) {
	    if($end0 >= $end1) {
		# ------------
		#     ----
		#
		# $hsp0 contains $hsp1

		$overlap = $end1 - $start1 + 1;
		$dl = $start1 - $start0;
		$dr = $end0 - $end1;
		$type = 0;
	    }
	    else {
		# --------
		#     --------
		#
		# the end of $hsp0 overlaps with the beginning of $hsp1

		$overlap = $end0 - $start1 + 1;
		$dl = $start1 - $start0;
		$dr = $end1 - $end0;
		$type = 1;
	    }
	}
	else {
	    # ----
	    #       ----
	    #
	    # $hsp0 is before $hsp1
	    $overlap = 0;
	    $dl = $end0 - $start0 + 1;
	    $dr = $end1 - $start1 + 1;
	    $type = 2;
	}
    }
    else {
	if($start0 <= $end1) {
	    if($end0 <= $end1) {
		#     ----
		# ------------
		#
		# $hsp0 is contained by $hsp1
		$overlap = $end0 - $start0 + 1;
		$dl = $start0 - $start1;
		$dr = $end1 - $end0;
		$type = 3;
	    }
	    else {
		#     --------
		# --------
		#
		# the beginning of $hsp0 overlaps with the end of $hsp1
		$overlap = $end1 - $start0 + 1;
		$dl = $start0 - $start1;
		$dr = $end0 - $end1;
		$type = 4;
	    }
	}
	else {
	    #       ----
	    # ----
	    #
	    # $hsp0 is after $hsp1
	    $overlap = 0;
	    $dl = $end1 - $start1 + 1;
	    $dr = $end0 - $start0 + 1;
	    $type = 5;
	}
    }

    return($overlap, $dl, $dr, $type);
}

sub to_range {
    my(@posns) = @_; # assumes they are ordered

    my $ranges;
    my $range;
    my $posn;

    #print "@posns\n";

    $ranges = [];
    if(@posns > 0) {
	$range = {'start' => $posns[0], 'end' => $posns[0]};
	push @{$ranges}, $range;
	if(@posns > 1) {
	    foreach $posn (@posns[1..$#posns]) {
		if($posn == $range->{'end'} + 1) {
		    $range->{'end'} = $posn;
		}
		elsif($posn > $range->{'end'}) {
		    $range = {'start' => $posn, 'end' => $posn};
		    push @{$ranges}, $range;
		}
		else {
		    TransDB::Root::warn("Error: to_range: positions are not in numerical order.");
		    return undef;
		}
	    }
	}
    }

    return $ranges;
}

sub ranges2string {
    my($ranges) = @_;

    my @strs;
    my $range;
    my $str;

    @strs = ();
    if(@{$ranges} > 0) {
	foreach $range (@{$ranges}) {
	    push @strs, sprintf("%d..%d", $range->{'start'}, $range->{'end'});
	}
	$str = join(',', @strs);
    }
    else {
	$str = '';
    }

    return $str;
}

sub aln_muscle {
    my($fn_aln, $id1, $seq1, $id2, $seq2) = @_;

    my $cmd;
    my $fh;

    $cmd = "muscle -clwstrict -quiet > $fn_aln 2> /dev/null";
    if(!open($fh, "| $cmd")) {
	warn "Error: aln_muscle: cannot open pipe to '$cmd'.";
	return 0;
    }
    printf $fh ">%s\n%s\n>%s\n%s\n", $id1, $seq1, $id2, $seq2;
    close($fh);

    return 1;
}

sub aln_hmmalign {
    my($fn_aln, $fn_pfam, $id1, $seq1, $id2, $seq2) = @_;

    my $cmd;
    my $fh;
    my $fn_aa;
    my $stat;
    my @L;
    my $i;
    my $id;
    my $seq;
    my $space;
    my $fn_aln2;
    my $in;
    my $out;
    my $aln;

    $fn_aa = $fn_aln . '.aa';
    $fn_aln2 = $fn_aln . '.psiblast';
    if(!open($fh, ">$fn_aa")) {
	warn "Error: aln_hmmalign: cannot open '$fn_aa' file for writing.";
	return 0;
    }
    printf $fh ">%s\n%s\n>%s\n%s\n", $id1, $seq1, $id2, $seq2;
    close($fh);

    $cmd = "hmmalign --outformat psiblast -o $fn_aln2 $fn_pfam $fn_aa";
    system($cmd);
    $stat = system($cmd);
    $stat >>= 8;
    if($stat != 0) {
	# FIXME - deal with errors
    }

    if(1) { # convert from 'PSIBLAST' to clustalw
	if(!open($fh, $fn_aln2)) {
	    warn "Error: aln_hmmalign: cannot open '$fn_aln2' file for reading.";
	    return 0;
	}
	@L = <$fh>;
	close($fh);

	if(!open($fh, ">$fn_aln")) {
	    warn "Error: aln_hmmalign: cannot open '$fn_aln' file for writing.";
	    return 0;
	}
	printf $fh "CLUSTAL W (1.81) multiple sequence alignment\n\n";
	for($i = 0; $i < @L; $i++) {
	    if(($L[$i] =~ /^CLUSTAL W/) or ($L[$i] =~ /\A\S*\Z/)) {
		print $fh $L[$i];
	    }
	    elsif($L[$i] =~ /\A(\s*)(\S+)(\s*)(\S+)/) {
		$space = $1 . $3;
		$id = $2;
		$seq = $4;

		$seq =~ s/\./-/g;
		$seq =~ tr/[a-z]/[A-Z]/;

		print $fh "$id$space$seq\n";
	    }
	    else {
		print $fh $_;
	    }
	}
	close($fh);
    }

    if(0) { # convert from 'A2M' (which looks like FastA) to clustalw
	system("aconvert -in f < $fn_aln2 > $fn_aln"); # FIXME - trap errors

	# now uppercase sequences and use only one gap char.
	# could do this with Bioperl but was complaining on the cluster
	if(!open($fh, $fn_aln)) {
	    warn "Error: aln_hmmalign: cannot open '$fn_aln' file for reading.";
	    return 0;
	}
	@L = <$fh>;
	close($fh);

	if(!open($fh, ">$fn_aln")) {
	    warn "Error: aln_hmmalign: cannot open '$fn_aln' file for writing.";
	    return 0;
	}
	for($i = 0; $i < @L; $i++) {
	    if(($L[$i] =~ /^CLUSTAL W/) or ($L[$i] =~ /\A\S*\Z/)) {
		print $fh $L[$i];
	    }
	    elsif($L[$i] =~ /\A(\s*)(\S+)(\s*)(\S+)/) {
		$space = $1 . $3;
		$id = $2;
		$seq = $4;

		$seq =~ s/\./-/g;
		$seq =~ tr/[a-z]/[A-Z]/;

		print $fh "$id$space$seq\n";
	    }
	    else {
		print $fh $_;
	    }
	}
	close($fh);
    }

    if(0) { # convert from stockholm to clustalw format that i2 can use
	$in = Bio::AlignIO->new('-file' => $fn_aln2, '-format' => 'Stockholm');
	$out = Bio::AlignIO->new('-file' => ">$fn_aln", '-format' => 'clustalw');
	$out->force_displayname_flat(1); # don't want Bioperl to add starts and stops
	while($aln = $in->next_aln) {
	    $aln->uppercase();
	    $aln->map_chars('\.', '-');
	    $out->write_aln($aln);
	}
    }

    if(0) { # convert hmmalign v2's weird clustal format
	if(!open($fh, $fn_aln)) {
	    warn "Error: aln_hmmalign: cannot open '$fn_aln' file for reading.";
	    return 0;
	}
	@L = <$fh>;
	close($fh);

	if(!open($fh, ">$fn_aln")) {
	    warn "Error: aln_hmmalign: cannot open '$fn_aln' file for writing.";
	    return 0;
	}
	for($i = 0; $i < @L; $i++) {
	    if(($L[$i] =~ /^CLUSTAL W/) or ($L[$i] =~ /\A\S*\Z/)) {
		print $fh $L[$i];
	    }
	    elsif($L[$i] =~ /\A(\s*)(\S+)(\s*)(\S+)/) {
		$space = $1 . $3;
		$id = $2;
		$seq = $4;

		$seq =~ s/\./-/g;
		$seq =~ tr/[a-z]/[A-Z]/;

		print $fh "$id$space$seq\n";
	    }
	    else {
		print $fh $_;
	    }
	}
	close($fh);
    }

    return 1;
}

sub mysystem {
    # http://www.perlmonks.org/?node_id=197500
    #
    # Use local signal handler so global handler
    # does not result in bad values in $? and $!

    local $SIG{CHLD} = '';
    return(system(@_));
}

1;
