#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

/* INTERPRETS 
 * Interaction prediction by Teriary Structure
 * Version 2.4 beta
 *
 * Version 2 series coded by Rob Russell, EMBL, Heidelberg
 *
 * Based on:
 * P. Aloy, R.B. Russell, 
 * Interrogating protein interaction networks through structural biology. 
 * Proc. Natl. Acad. Sci. USA, 99, 5896-5901, 2002. 
 * PubMed ID 11972061
 *
 * Modifications to this approach (-mode 3)
 * 1. Score now considered as difference from template
 *   Values of Z = 0 implies no change from the template
 *   Values above are better, below are worse than template
 *
 * 2. Randomisations changed (-mode 3)
 *  Now the idea is to mutate randomly the sequence rather
 *   than randomise - down to the identity between the query proteins
 *   and their template components
 *
 * 3. Randomisations changed (-mode 4)
 *  As for 2, but we (Matt, Damien, Chad, Rob) deduced that a better random
 *   model is to choose a fixed low identity - perhaps based on Aloy et al JMB 2003
 *   that is - 30% identity is when things more often don't interact than do
 *   The problem with 2. is that 100% identical things
 *   Note that we are always using the score - ref_score approach - that is the difference
 *   rather than the raw.  A la FoldX logic
 *
 * 4. Insertion penalty - lowest value in the matrix to be added for every residue inserted +/- 1 to 
 *    residue involved in interaction. - NOT DONE YET
 *
 * 5. mode 0 added - output fasta that the program is using - probably with stamp, etc.
 *  - NOT DONE YET
 */

#define MAX_ID_LEN 100
#define MAX_SEQ_LEN 10000
#define PRECIS 1000
#define angle 0.942477
#define distance 1.54
#define a3to1_acids3 "ALA ARG ASN ASP CYS GLN GLU GLY HIS ILE LEU LYS MET PHE PRO SER THR TRP TYR VAL ASX GLX UNK CYH HET"
#define a3to1_acids1 "ARNDCQEGHILKMFPSTWYVBZXch"
#define AA1a "A_CDEFGHI_KLMN_PQRST_VW_Y_"

static float AA_ABUN[26] =
 /* Ala,  ---,  Cys,  Asp,  Glu,  Phe,  Gly,  His,  Ile,  ---,  Lys,  Leu,  Met */
 {  8.5,  0.0,  1.0,  5.8,  5.7,  3.6,  8.5,  2.6,  5.3,  0.0,  6.0,  8.1,  2.0,  
 /* Asn,  ---,  Pro, Gln, Arg, Ser, Thr, ___, Val, Trp, ---, Tyr, --- */
    4.7,  0.0,  4.7, 3.5, 4.2, 6.9, 6.2, 0.0, 7.7, 1.3, 0.0, 3.6, 0.0 };

struct parameters {
  int mode;
  int quiet;
  int disp;
  int verbose;
  int really_verbose;
  int nc_min;
  int max_hb;
  int max_sb;
  int max_vdw;
  int nrand;
  int rand;
  int show_template;
  int diff;
  int onepair;
  char aln_file1[1000];
  char aln_file2[1000];
  char dom_file[1000];
  char ras_file[1000];
  char ats_file[1000];
  char ss_mat_file[1000];
  char sm_mat_file[1000];
  char p1[1000];
  char p2[1000];
  char pdb[10];
  float id_min;
  float vdw_tol;
  float sb_tol;
  float hb_tol;
  float ca_ca_tol;
};

void exit_error() {
   fprintf(stderr,"format: i2 -a <afile 1> <afile 2> -d <domfile>\n");
   fprintf(stderr,"        i2 -help gives details of all commands\n");
   exit(-1);
}
void i2_help(struct parameters *par) {
   fprintf(stderr,"format:  i2 -a <afile 1> <afile 2> -d <domfile>\n");
   fprintf(stderr," options & parameters:\n");
   fprintf(stderr,"  -q                     = quiet mode (no comments)\n");
   fprintf(stderr,"  -fasta                 = just output two fasta formatted sequence & quit\n");
   fprintf(stderr,"  -count                 = just output contacts & quit (for parameter derivation)\n");
   fprintf(stderr,"  -hb_tol <float>        = H-bond atom-atom distance tolerance for contact (def. %3.1f)\n",par->hb_tol);
   fprintf(stderr,"  -sb_tol <float>        = salt-bridge atom-atom distance tolerance for contact (def. %3.1f)\n",par->sb_tol);
   fprintf(stderr,"  -vdw_tol <float>       = VDW atom-atom distance tolerance for contact (def. %3.1f)\n",par->vdw_tol);
   fprintf(stderr,"  -ca_ca_tol <float>     = calpha-calpha distance for speed filter (def. %4.1f)\n",par->ca_ca_tol);
   fprintf(stderr,"  -nc_min <int>          = min no. of contacts (within tolerances) to include res-res contact (def: %d)\n",par->nc_min);
   fprintf(stderr,"  -ss_mat <file>         = file for side- to side-chain pair-potentials (def. %s)\n",par->ss_mat_file);
   fprintf(stderr,"  -sm_mat <file>         = file for side- to main-chain potentials (def. %s)\n",par->sm_mat_file);
   fprintf(stderr,"  -rand <int>            = number of randomisations to do\n");
   fprintf(stderr,"  -diff                  = use difference score (i.e. from template)\n");
   fprintf(stderr,"  -op <string> <string>  = only consider single pair\n");
   fprintf(stderr,"  -mode <1,2,3,4>        = randomisation method used\n");
   fprintf(stderr,"            mode 1       = interprets paper (PNAS 2002)\n");
   fprintf(stderr,"            mode 2       = interprets paper, but randomising model sequences (PNAS 2002) - quite silly\n");
   fprintf(stderr,"            mode 3       = randomisations done according to sequence identities A-A'/B-B'\n");
   fprintf(stderr,"            mode 4       = randomisations done according to low sequence identity (JMB 2003)\n");
   fprintf(stderr,"  -show_template         = report template domain pair in output\n");
   fprintf(stderr,"  -help or -h            = this summary\n");
   exit(-1);
}

struct astruc {
  int n;
  int len;
  char **id;
  char **s;
};

struct pp_mat {
    float *m;
    float **p;
};

struct brookn { /* structure to represent brookhaven residue numbers */
    int n;      /* numerical part of number */
    char cid;   /* chain identifier */
    char in;    /* insertion code */
};

struct domain_loc {              /* This structure allows rather complex domains to be described */
   char filename[100];
   char id[100];
   int nobj;                    /* The number of objects considered within the named file */
   int *type;                   /* The type that each object is:
                                        0 ==> an error
                                        1 ==> All of the residues in the file
                                        2 ==> A particular chain
                                        3 ==> A particular named region (eg. A 25 _ to B 10 _ ) */
   struct brookn *start;        /* There will be a start and end for each 'object' */
   struct brookn *end;
   int *reverse;                        /* if 1, then reverse invert the object N to C */
   int **coords;
   char *aa;
   char *align;
   char *oldalign;
   char *sec;
   struct brookn *numb;
   int ncoords;
   int *use;
   float **R;                   /* Initial transformation */
   float *V;
   float **r;                   /* current transformation, when STAMP is done, we must update(r,v,R,V) to get the final transfor
mation */
   float *v;
   float value;
};

struct atom_details {
    char **orders;
    int  **use;
    int *functional;
    int *functional_start;
    int *found;
};

struct resdat {
   int pos;   /* pointer to native sequence */
   char aa;  /* Amino acid type */
   char aag; /* aa if no grouping, new symbol if grouping (i.e. a, b, c...etc.) */
   struct brookn n; /* PDB numbering */
   int use;
};

struct residue {
   char aa;
   char ss;
   int acc;
   int n;
   int nats;
   struct brookn pdbnum;
   int **co;
   int *oc;
   int *t; /* 0 = Carbon or Sulfur; 1 = Nitrogen; 2 = Oxygen */
};

struct contacts {
   int r_i,r_j;
   int a_i,a_j;
   float d;
   int type;
};

char ltou(char c) {

	if (c>='a' && c<='z')
	   return c - 'a' + 'A';
	else return c;
}
char utol(char c) {

	if (c>='A' && c<='Z')
	   return c - 'A' + 'a';
	else return c;
}

void closefile(FILE *handle, char *filename) {


	char b[1000];
	if((strcmp(&filename[strlen(filename)-2],".Z")==0) || 
           (strcmp(&filename[strlen(filename)-3],".gz")==0)) { 

	    /* This fgets prevents "Broken pipe" messages from appearing when running under Solaris,
	     *  and silly newlines to stderr when running under OSF - something about a wrong implementation
	     *  of pclose()? */
	    while(fgets(b,999,handle)!=NULL);
	    pclose(handle);
	} else { 
	    fclose(handle);
	}
/*	printf("Closed file %s\n",filename); */
}

FILE *openfile(char *filename, char *type) {

	FILE *handle;
	char command[1000];

/*	printf("Attempting to open file %s\n",filename); */
	if(strcmp(type,"r")==0) { /* A read file */
		if(strcmp(&filename[strlen(filename)-2],".Z")==0) { /* UNIX compression */
/*		     sprintf(command,"zcat %s 2> /dev/null",filename);  */
		     sprintf(command,"zcat %s",filename);  
		     handle=popen(command,"r");
		} else if(strcmp(&filename[strlen(filename)-3],".gz")==0) { /* gzipped  */
/*		     sprintf(command,"gunzip -c %s 2> /dev/null",filename);  */
		     sprintf(command,"gunzip -c %s",filename); 
		     handle=popen(command,"r");
		} else { /* presume no compression */
		    handle=fopen(filename,"r");
		}
	} else {
	  	    handle=fopen(filename,"w");
	}
/*	printf("Opened file %s\n",filename); */
	return handle;
}


char *getfile(char *code_safe, char *dirfile, int code_length, FILE *OUTPUT) {
  /* getfile: 
   *  opens a file containing a list of appropriate directories, suffixes and
   *  prefixes.  It then makes combinates of these with the supplied code
   *  to see if a file exists 
   *
   * Allocates memory for a string corresponding to the filename read in,
   *  then returns this string or '\0' if none is found 
   *
   * Modification March 1999 - now tries to expand into a "distr" style hierarchy */

	char *temp;
	char *dir;
	char *buff;
	char prefix[10],suffix[10];
	char *code;
	char *CODE;
	char *two_letter; 
	char a;

	int i,j;
	int found;

	FILE *DIR;

	dir=(char*)malloc(1000*sizeof(char)); 
	buff=(char*)malloc(1000*sizeof(char)); 
	temp=(char*)malloc(1000*sizeof(char)); 
	CODE=(char*)malloc((strlen(code_safe)+1)*sizeof(char));
	code=(char*)malloc((strlen(code_safe)+1)*sizeof(char));
	two_letter = (char*)malloc(3*sizeof(char));

	/* copy code_safe into code to prevent the changing case from
	 *  carrying on through the program */

	a = code[code_length];
	code[code_length]='\0'; /* truncates the code if necessary */
	for(i=0; i<strlen(code_safe); ++i) code[i]=utol(code_safe[i]);
	for(i=0; i<strlen(code_safe); ++i) CODE[i]=ltou(code_safe[i]); /* upper case code */
	code[code_length]='\0';
	CODE[code_length]='\0';
	two_letter[0] = code[1]; two_letter[1] = code[2]; two_letter[2] = '\0'; /* For PDB distr style */

	if((DIR=fopen(dirfile,"r"))==NULL) {
	   fprintf(stderr,"error: no directories file found\n");
	   temp[0]='\0';
	   return temp;
	}
	found=0;
	while((fgets(buff,10000,DIR))!=NULL) {

	  sscanf(buff,"%s %s %s",dir,&prefix[0],&suffix[0]);
	  if(prefix[0] == '_') { prefix[0] = '\0'; }
	  if(suffix[0] == '_') { suffix[0] = '\0'; }

	  /* Lower case, all files one directory */
	  sprintf(temp,"%s/%s%s%s",dir,prefix,code,suffix);
	  if(testfile(temp)) { found=1; break; } 

	  /* Lower case, files in 'distr' type directories */
	  sprintf(temp,"%s/%s/%s%s%s",dir,two_letter,prefix,code,suffix);
	  if(testfile(temp)) { found=1; break; } 

	  /* Upper case, all files one directory */
	  sprintf(temp,"%s/%s%s%s",dir,prefix,CODE,suffix);
	  if(testfile(temp)) { found=1; break; } 

	  /* Upper case, files in 'distr' type directories */
	  sprintf(temp,"%s/%s/%s%s%s",dir,two_letter,prefix,CODE,suffix);
	  if(testfile(temp)) { found=1; break; } 

	}

	if(found==0) {
	   temp[0]='\0';
	}
	fclose(DIR);
	free(dir); 
	free(buff);
	free(code);
	free(CODE);
	free(two_letter);

	return temp;
}

int print_vector(float *V) {
	int i;
	printf("(");
	for(i=0; i<3; ++i) {
	   printf("%8.5f",V[i]);
	   if(i<2) printf(", ");
	}
	printf(")");
	printf(" length: %8.5f\n",
	   sqrt(V[0]*V[0]+V[1]*V[1]+V[2]*V[2]));
}
int vector_unify(float *V) {

	int i;
	float magnitude;

	magnitude=sqrt(V[0]*V[0]+V[1]*V[1]+V[2]*V[2]);
	for(i=0; i<3; ++i) 
	   V[i]=((1/magnitude)*(float)V[i]);
	
	return 0;
}

float *vector_diff(float *V1, float *V2) {
	int i;
	float *V3;

	V3=(float*)malloc(3*sizeof(float));

	for(i=0; i<3; ++i) {
	   V3[i]=V1[i]-V2[i];
	}

	return V3;
}

float *vector_ave(float *V1, float *V2) {
	int i;
	float *V3;

	V3=(float*)malloc(3*sizeof(float));

	for(i=0; i<3; ++i) {
	   V3[i]=((float)(V1[i]+V2[i])/2);
	}

	return V3;
}

float *vector_cross(float *V1, float *V2) {

	int i;
	float *V3;

/* SMJS Removed V3=V3 */
	V3=(float*)malloc(3*sizeof(float));

	V3[0]=V1[1]*V2[2]-V1[2]*V2[1];
	V3[1]=V1[2]*V2[0]-V1[0]*V2[2];
	V3[2]=V1[0]*V2[1]-V1[1]*V2[0];

	return V3;
}

float *vector_set_dist(float *V1, float R) {

	/* given a vector and a distance, update the vector to have a new length */
	int i;
	float *V2;
	float r;

	V2=(float*)malloc(3*sizeof(float));
	r=sqrt(V1[0]*V1[0]+V1[1]*V1[1]+V1[2]*V1[2]);
	for(i=0; i<3; ++i) 
	   V2[i]=(R/r)*V1[i];

	return V2;
}

char **c_split(char *str, int *n,  char delimiter) {
	/* breaks a string up into substrings using a delimiter */

	int i,j,k,len;
	char **sstr;

	sstr=(char**)malloc(sizeof(char*));
	
	(*n)=0;
	i=0;
	len=strlen(str);

 	while(i<len && str[i]!='\0' && str[i]!='\n') {
	   while(i<len && str[i]==delimiter) i++;
	   if(i>=len || str[i]=='\0' || str[i]=='\n') break;
	   /* we are at a new sub-string */
	   sstr[(*n)]=(char*)malloc(sizeof(char));
	   j=0;
	   while(i<len && str[i]!='\0' && str[i]!='\n' && str[i]!=delimiter) {
		sstr[(*n)][j]=str[i];
		j++; i++;
	        sstr[(*n)]=(char*)realloc(sstr[(*n)],(j+1)*sizeof(char));
	   }
	   sstr[(*n)][j]='\0';
	   (*n)++;
	   sstr=(char**)realloc(sstr,((*n)+1)*sizeof(char*));
	   i++;
	}
/*	printf("String %s breaks into %d sub-strings delimited by '%c' characters\n",
		str,(*n),delimiter);
	for(i=0; i<(*n); ++i) printf("%s, ",sstr[i]);
	printf("\n"); 
*/
	return sstr;
}

struct pp_mat *read_pp_mat(FILE *SS, FILE *SM, struct parameters *par) {

   int i,j;
   int nt;
   int *p;
   int len;

   float tmp;

   char buff[1000];
   char **t;
   struct pp_mat *pp;

   len = strlen(AA1a);

   p = (int*)malloc(len*sizeof(int));
   pp = (struct pp_mat*)malloc(sizeof(struct pp_mat));
   pp->m = (float*)malloc(len*sizeof(float));
   pp->p = (float**)malloc(len*sizeof(float*));
   for(i=0; i<len; ++i) {
     pp->p[i] = (float*)malloc(len*sizeof(float));
   }

   if(par->verbose==1) { printf("# Side-chain to side-chain potential matrix\n"); }
   while(fgets(&buff[0],999,SS)!=NULL) {
      t = c_split(&buff[1],&nt,' ');
      if(buff[0]=='#') { /* Index line */
        if(par->really_verbose) {
           printf("%s",buff);
        }
        if(par->verbose) { printf("#"); }
        for(i=0; i<nt; ++i) {
          p[i]=-1;
          if(par->verbose==1) {
              printf(" %6s",t[i]);
          }
          for(j=0; j<len; ++j) {
            if(t[i][0] == AA1a[j]) {
               p[i] = j;
               break;
            }
          }    
          if(p[i] == -1) {
            fprintf(stderr,"ERROR: bad amino acid %c in side-chain side-chain %s",t[i][0],buff);
            exit(-1);
          }
        }
        if(par->verbose==1) { printf("\n"); }
        i=0;
      } else {
        if(par->verbose==1) { printf("#%c ",AA1a[p[i]]); }
        for(j=0; j<nt; ++j) {
          sscanf(t[j],"%f",&tmp);
          pp->p[p[i]][p[j]]=tmp;
          pp->p[p[j]][p[i]]=tmp;
          if(par->verbose) {
            printf("%6.3f ",pp->p[p[i]][p[j]]);
          }
        }
        if(par->verbose==1) { printf("\n"); }
        i++;
      } 
      for(j=0; j<nt; ++j) { free(t[j]); }
      free(t);
   }
   if(par->verbose==1) { printf("# Side-chain to main-chain potential matrix\n"); }
   while(fgets(&buff[0],999,SM)!=NULL) {
     t = c_split(buff,&nt,' ');
     if(nt!=2) {
       fprintf(stderr,"ERROR: bad line in side-chain main-chain file %s",buff);
       exit(-1);
     }
     p[0]=-1;
     for(i=0; i<strlen(AA1a); ++i) {
       if(AA1a[i] == t[0][0]) {
          p[0] = i;
          break;
       }
     }
     if(p[0] == -1) {
       fprintf(stderr,"ERROR: bad amino acid %c in side-chain main-chain %s",t[0][0],buff);
       exit(-1);
     }
     sscanf(t[1],"%f",&tmp);
     pp->m[p[0]] = tmp;
     if(par->verbose) {
       printf("#%s %2d %8.3f\n",t[0],p[0],pp->m[p[0]]); 
     }

     for(i=0; i<nt; ++i) { free(t[i]); }
     free(t);
   }
   free(p);
   return pp;
}

struct atom_details *get_atom_details(FILE *IN) {
   
   int i,j,k;
   int ntokens;
   int aa_len;
   int pt;

   struct atom_details *details;

   char buff[1000];
   char aa;
   char **t;

   aa_len = strlen(AA1a);
   details = (struct atom_details*)malloc(sizeof(struct atom_details));
   details->orders=(char**)malloc(aa_len*sizeof(char*));
   details->use  =(int**)malloc(aa_len*sizeof(int*));
   details->functional=(int*)malloc(aa_len*sizeof(int));
   details->functional_start=(int*)malloc(aa_len*sizeof(int));
   details->found=(int*)malloc(aa_len*sizeof(int));
   for(i=0; i<aa_len; ++i) {
       details->orders[i]=(char*)malloc(100*sizeof(char));
       details->use[i]=(int*)malloc(20*sizeof(int));
       details->found[i]=0; /* will be 1 when values set */
       for(j=0; j<16; ++j) {
           details->use[i][j]=0;
           strncpy(&details->orders[i][j*4],"____",4);
       }
   }
 
   while(fgets(&buff[0],299,IN)!=NULL) {
     if((strncmp(buff,"Use ",4)==0) || (strncmp(buff,"Atom",4)==0)) {
/*       printf("%s\n",buff); */
       t = c_split(&buff[0],&ntokens,' ');
       if(ntokens!=18) {
          fprintf(stderr,"Error reading orders file.  Last line was %s\n",buff);
          exit(-1);
       }

/*
       printf("String %s split into %4d tokens\n",buff,ntokens);
       for(i=0; i<ntokens; ++i) { printf("%4d %s\n",i,t[i]); }
*/

       aa = t[1][0];
       pt = (int)(aa-'A');
       if(strncmp(buff,"Atom",4)==0) {
             for(j=0; j<16; ++j) {
                 for(k=0; k<strlen(t[2+j]); ++k) {
                    if(t[2+j][k]=='_') { t[2+j][k]=' '; }
                 }
                 strncpy(&details->orders[pt][j*4],t[2+j],4);
             }
       }
       if(strncmp(buff,"Use ",4)==0) {
             for(j=0; j<16; ++j) {
                 details->use[pt][j]=(int)(t[j+2][0]-'0');
                 if(details->use[pt][j]==2) {
                      details->functional[pt]=j;
                 }
             }
       }
       for(i=0; i<ntokens; ++i) { free(t[i]); }
       free(t);
     }
   }
   return details;
}

float *build_cb(float *GLY_N, float *GLY_CA, float *GLY_CO, FILE *OUT) {
	int i;
	float *V1,*V2;
	float *average;
	float *cross;
	float *cb,*GLY_CB;

	cb=(float*)malloc(3*sizeof(float));

	/* this sets the origin to the position of GLY_CA */
	V1=vector_diff(GLY_N,GLY_CA);
	V2=vector_diff(GLY_CO,GLY_CA);

	/* set lengths to unity */
	vector_unify(V1);
	vector_unify(V2);

	/* get average of V1 and V2 */
	average=vector_ave(V1,V2);
	/* reverse the sign to make it point in the opposite direction */
	for(i=0; i<3; ++i) average[i]=-1*average[i]; 
	vector_unify(average);

	/* cross product */
	cross=vector_cross(V1,V2); 
	vector_unify(cross);

	/* find coordinates based on ideal angle */
	for(i=0; i<3; ++i) {
	  cb[i]=cross[i]*(float)cos((double)angle)+average[i]*(float)cos((double)angle);
	}

	/* length of cb is unity, set to ideal */
	GLY_CB=vector_set_dist(cb,distance);

	/* this vector treats CA as the origin, we must
	 * get the "real" coordinate back */
	


	/* done, return the result */
	for(i=0; i<3; ++i)  {
	  GLY_CB[i]=GLY_CB[i]+GLY_CA[i];
        }

	free(cb);
	return GLY_CB;


}

int *ibuild_cb(int *iGLY_N, int *iGLY_CA, int *iGLY_CO,  FILE *OUT) {
      /* Builds ghost CB */

      int i;
      int *iCB;

      float *GLY_N,*GLY_CA,*GLY_CO;
      float *CB;

      iCB    =(int*)malloc(3*sizeof(int));
      GLY_N  =(float*)malloc(3*sizeof(float));
      GLY_CA =(float*)malloc(3*sizeof(float));
      GLY_CO =(float*)malloc(3*sizeof(float));

      for(i=0; i<3; ++i) {
          GLY_N[i]  = (float)iGLY_N[i]/(float)PRECIS;
          GLY_CA[i] = (float)iGLY_CA[i]/(float)PRECIS;
          GLY_CO[i] = (float)iGLY_CO[i]/(float)PRECIS;
      }
      CB = build_cb(GLY_N,GLY_CA,GLY_CO,stdout);

      for(i=0; i<3; ++i) {
         iCB[i] = (int)((float)PRECIS*CB[i]);
      }
      free(CB);
      return iCB;
}
      
int getpointer(char *atomname, char aa, char **atom_orders) {
   /* ATOM     21  OG  SER     3       5.350  10.005  19.941  1.00 11.98 
    * 01234567890123456 */
   /*  " N   CA  C   O   CB  CG1 CG2 CD1--------------------------------", */
   /*   0123456789012345678901234567890123456789012345678901234567890123456789 */
   /*   |         1         2         3         4         5         6  | */

	int pointer;
  	char *tmpstring;

 	tmpstring = strstr(atom_orders[(int) (aa - 'A')], atomname);

	if(tmpstring != NULL) { 
	    pointer = 64 - strlen(tmpstring);
	    pointer = pointer / 4;
	}
	else {
	    pointer = -1;
	}
	return pointer;

}

char a3to1(char *a3) {
/* Converts three letter amino acid code to one letter
 *  amino acid code 
 * Note: CYS refers to cystine, CYH refers to cysteine */
	int i;
	char new;

	new='X';
	for(i=0;i<strlen(a3to1_acids1); i++) {
	   if (strncmp(&a3to1_acids3[i*4],a3,3) == 0) 
	      new=(char)a3to1_acids1[i];
	}   /* end of for  */

	return(new);
} 
struct residue *getallcoords(struct domain_loc *domain, int *nres, int *natoms, char *pdb_files, char **atom_orders, int verbose, int really_verbose) {

/* This reads in all atoms for a set of residues specified by a
 *  domain descriptor.  See the end of file for a sort of historical
 *  derivation of how the amino acids should/could match */



	int i,j,k;
	int pointer;
	int begin,end;
	int number;
	int onlastresidue;

	char *pdbfile;
	char buff[100];
	char lastres[20];
	char atomname[10];
	char tmpstring[100];
	char cid,in,alt;

	float x;

	struct residue *res;

	FILE *PDB;
	

	(*nres)   = -1;
	(*natoms) = 0;

	if((PDB=openfile(domain[0].filename,"r"))==NULL) {
             pdbfile=getfile(domain[0].id,pdb_files,4,stdout);
             if(pdbfile[0]=='\0') {
                 fprintf(stderr,"Error in readats() no PDB file found for %s\n",domain[0].id);
                 return NULL;
             }
             strcpy(domain[0].filename,pdbfile);
             free(pdbfile);
        }

	if(verbose) printf("# Filename is %s\n",domain[0].filename);

	res = (struct residue*)malloc(sizeof(struct residue));
	for(i=0; i<domain[0].nobj; ++i) { 
            if(verbose) {
		printf("# Domain %s object %4d type %4d\n",domain[0].id,i,domain[0].type[i]); 
	   
                switch(domain[0].type[i]) {
                     case 1: printf("#  all residues"); break; 
                     case 2: printf("#  chain %c",domain[0].start[i].cid); break;
                     case 3: printf("#  from %c %4d %c to %c %4d %c",
                         domain[0].start[i].cid,domain[0].start[i].n,domain[0].start[i].in,
                         domain[0].end[i].cid,domain[0].end[i].n,domain[0].end[i].in); break;
                 } 
                 printf("\n");
	    }

	   closefile(PDB,domain[0].filename); 
	   PDB=openfile(domain[0].filename,"r");
	   begin=0;  lastres[0]='\0'; onlastresidue=0;
	   while(fgets(buff,99,PDB)!=NULL) {
	     if((strncmp(buff,"ENDMDL",6)==0 || strncmp(buff,"END   ",6)==0) && begin==1) { 
/*			printf("Ending at %s\n",buff); */
			break; 
	     } 
	     if(strncmp(buff,"ATOM  ",6)==0 && strncmp(&buff[17],"ACE",3)!=0 && strncmp(&buff[17],"FOR",3)!=0) {
	      /* get chain, number and insertion code */
	      cid=buff[21]; sscanf(&buff[22],"%d",&number); in=buff[26]; alt=buff[16]; /* alternate position indicator */
	      /* Need to fix this */
              /* Fixed I think - I was stupid and this was at the end, not the begining where it belongs */
	      if(begin && 
		 ((domain[0].type[i]==2 && domain[0].start[i].cid!=cid) ||
		  (strncmp("ATOM  ",buff,6)!=0))) {
		    onlastresidue=1;
	      }
/*	      printf("%s cid = %c n = %d in = %c compared to %c %d %c \n",buff,cid,number,in,
			domain[0].start[i].cid,domain[0].start[i].n,domain[0].start[i].in); */
	      if(!begin && 
		 ((domain[0].start[i].cid==cid && domain[0].start[i].n==number && 
		   domain[0].start[i].in==in) ||  /* range */
		   (domain[0].start[i].cid==cid && domain[0].type[i]==2) || /* chain */
		   (domain[0].type[i]==1) ))  {  /* all residues */
			begin=1;
			onlastresidue=0;
/*			printf("Begining at %s\n",buff); */
	      }
	      /* Test if this is a new residue */
	      if((strncmp(lastres,&buff[17],11)!=0)) {
		if(begin) {
		  if(onlastresidue==1) break;
		  (*nres)++;
		  res = (struct residue*)realloc(res,((*nres)+1)*sizeof(struct residue));
		  res[(*nres)].n=(*nres);
		  res[(*nres)].oc=(int*)malloc(16*sizeof(int));
		  res[(*nres)].co=(int**)malloc(16*sizeof(int*));
		  res[(*nres)].t=(int*)malloc(16*sizeof(int));
	          for(j=0; j<16; ++j) res[(*nres)].oc[j]=0;
		  res[(*nres)].aa=a3to1(&buff[17]);
                  if(cid==' ') res[(*nres)].pdbnum.cid='_';
                  else res[(*nres)].pdbnum.cid=cid;
                  if(in==' ') res[(*nres)].pdbnum.in='_';
                  else res[(*nres)].pdbnum.in=in;
                  res[(*nres)].pdbnum.n=number;
		  res[(*nres)].nats=0;
		  res[(*nres)].acc=-1;
		  res[(*nres)].ss='u';
		}
	      }
              if((really_verbose==1) && (begin==1) && (strncmp(&buff[12]," N  ",4)==0)) {
		printf("Residue %4d %c %4d %c %c \n",(*nres),
			res[(*nres)].pdbnum.cid, res[(*nres)].pdbnum.n,
			res[(*nres)].pdbnum.in,res[(*nres)].aa);
	      }
	      strncpy(lastres,&buff[17],11); lastres[11]='\0';
	      if(begin && (alt==' ' || alt=='A' || alt=='1') ) { /* only read one of multiple ATOM entries */

		   strncpy(&atomname[0],&buff[12],4);	
		   atomname[4]='\0';

		   pointer = getpointer(atomname, res[(*nres)].aa, atom_orders);
		   if(pointer != -1 && pointer < 16) {
                      if((atomname[1] == 'C') || (atomname[1] == 'S')) { 
                           res[(*nres)].t[pointer] = 0;
                      } else if(atomname[1] == 'N') { 
                           res[(*nres)].t[pointer] = 1;
                      } else if(atomname[1] == 'O') { 
                           res[(*nres)].t[pointer] = 2;
                      } else {
                           res[(*nres)].t[pointer] = 3;
                      }
	  	      res[(*nres)].oc[pointer] = 1;
		      res[(*nres)].co[pointer] = (int *) malloc(3 * sizeof(int)); 

		      if(really_verbose==1)  {
                         printf(" atom %4d (%4d) %s occ %4d (%d)\n",
                          res[(*nres)].nats,pointer,atomname,res[(*nres)].oc[pointer],res[(*nres)].nats);
                      }

		      for(j = 0; j < 3; ++j) {
			  strncpy(&tmpstring[0], &buff[30 + j * 8], 8);
			  tmpstring[8] = '\0';
			  sscanf(&buff[30 + j * 8], "%f", &x);
			  res[(*nres)].co[pointer][j] = (int) (PRECIS * x);
                      }
		      (*natoms)++; res[(*nres)].nats++;
	              if((*nres)>MAX_SEQ_LEN) {
		        fprintf(stderr,"error: number of coordinates read surpasses memory limit\n");
		        return NULL;
	              }
		   } else {
		      printf("# WARNING: strange atom type in %s",buff);
                   }
                      
	      } 
	   } 
	   if(begin && 
	     (domain[0].type[i]==3 && 
	      domain[0].end[i].cid==cid && domain[0].end[i].n==number && 
	      domain[0].end[i].in==in && domain[0].type[i]==3)) {
		    onlastresidue=1;
	  }
	}
        }
	closefile(PDB,domain[0].filename); 
	if(!begin) {
	   fprintf(stderr,"error: begin of sequence not found in PDB file\n");
	   (*nres)=0; (*natoms)=0;
	   return NULL;
	} else {
	   (*nres)++;
	}

#if(0)
        /* Build CB for glycines */
        for(i=0; i<(*nres); ++i) {
          if((res[i].aa=='G') && (res[i].oc[0]==1) && 
             (res[i].oc[1]==1) && (res[i].oc[2]==1) && (res[i].oc[3]==1)) {
                res[i].co[4] = ibuild_cb(res[i].co[0],res[i].co[1],res[i].co[2],stdout);
                res[i].oc[4]=1;
                res[i].t[4]=0;
                res[i].nats++;
                if(verbose) { 
  		   printf("# Building CB for residue %4d %c %4d %c %c \n",
                        i,res[i].pdbnum.cid, res[i].pdbnum.n,res[i].pdbnum.in,res[i].aa);
                }
          }
        }
#endif


	return res;
}


float idist(int *atm1, int *atm2, int PRECISION) {
        /* Finds the square distance between to ints *atm1, *atm2 */
        int i;
        float D;

        D=0.0;
        for(i=0; i<3; ++i)
           D+=((float)(atm1[i]-atm2[i])/(float)PRECISION)*((float)(atm1[i]-atm2[i])/(float)PRECISION);
        return sqrt(D);
}

struct rdist {
  int i,j;
  int ii,jj;
  float d;
};

int compdist(const void *l, const void *r) {

   int i;

   const struct rdist *left = l;
   const struct rdist *right = r;


   if(right->d < left->d) {
      return 1;
   } else {
      return -1;
   }
}


int ***getcontacts(struct residue *r1, int nr1, struct residue *r2, int nr2, struct atom_details *atoms, struct parameters *par) {

    int i, j, ii, jj;
    int k, l;
    int n;
    int p1, p2;
    int type;
    int nwithin; /* just for testing ca_ca_tol */
    int contact;
    char name1[5];
    char name2[5];

    int *hb_count1, *hb_count2;
    int *sb_count1, *sb_count2;
    int *vdw_count1, *vdw_count2;
    int ***cont;

    float dist;

    struct rdist *D;

    hb_count1 = (int *) malloc(nr1 * sizeof(int));
    vdw_count1 = (int *) malloc(nr1 * sizeof(int));
    sb_count1 = (int *) malloc(nr1 * sizeof(int));

    hb_count2 = (int *) malloc(nr2 * sizeof(int));
    vdw_count2 = (int *) malloc(nr2 * sizeof(int));
    sb_count2 = (int *) malloc(nr2 * sizeof(int));
    
    for(i = 0; i < nr1; ++i) {
	hb_count1[i] = 0;
	vdw_count1[i] = 0;
	sb_count1[i] = 0;
    }
    for(i = 0; i < nr2; ++i) {
	hb_count2[i] = 0;
	vdw_count2[i] = 0;
	sb_count2[i] = 0;
    }

    /* Compute distances */

    D = (struct rdist *) malloc(sizeof(struct rdist));
    n = 0; 
    for(i = 0; i < nr1; ++i) {   
	for(j = 0; j < nr2; ++j) {
	    dist = idist(r1[i].co[1], r2[j].co[1], PRECIS);
	    if(dist <= par->ca_ca_tol) {
		for(ii = 0; ii < 16; ++ii) {
		    for(jj = 0; jj < 16; ++jj) {
			if((r1[i].oc[ii] == 1) && (r2[j].oc[jj] == 1)) {
			    D[n].i = i;
			    D[n].j = j;
			    D[n].ii = ii;
			    D[n].jj = jj;
			    D[n].d = idist(r1[i].co[ii], r2[j].co[jj], PRECIS);

			    /*
			    p1 = (int)(r1[i].aa - 'A'); strncpy(name1,&(atoms->orders[p1][ii*4]),4); name1[4]='\0';
			    p2 = (int)(r2[j].aa - 'A'); strncpy(name2,&(atom->orders[p2][jj*4]),4); name2[4]='\0';
			    printf("Read %3d %3d (n %3d) %c %4d %c (%c %4s) <=> %3d %3d (n %3d) %c %4d %c (%c %4s) =  %8.3f\n",
				   i,ii,r1[i].nats,r1[i].pdbnum.cid,r1[i].pdbnum.n,r1[i].pdbnum.in,r1[i].aa,name1,
				   j,jj,r2[j].nats,r2[j].pdbnum.cid,r2[j].pdbnum.n,r2[j].pdbnum.in,r2[j].aa,name2,
				   D[n].d);
			    */

			    n++;
			    D = (struct rdist *) realloc(D, (n + 1) * sizeof(struct rdist));
			}
		    } 
		}
	    }
	}
    }

    /* Sort distances */
    qsort((char *) D, n, sizeof(struct rdist), compdist);  

    cont = (int ***) malloc(nr1 * sizeof(int **));
    for(i = 0; i < nr1; ++i) {
	cont[i] = (int **) malloc(nr2 * sizeof(int *));
	for(j = 0; j < nr2; ++j) {
	    cont[i][j] = (int *) malloc(10 * sizeof(int));
	    for(k = 0; k < 10; ++k) {
		cont[i][j][k] = 0;
	    }
	}
    }

    nwithin = 0;
    for(k = 0; k < n; ++k) {
	i    = D[k].i;
	j    = D[k].j;
	ii   = D[k].ii;
	jj   = D[k].jj;
	dist = D[k].d;
	contact = 0;

	if((dist > par->hb_tol) && (dist > par->sb_tol) && (dist > par->vdw_tol)) {
	    break;
	}

	if((r1[i].oc[ii] == 1) && (r2[j].oc[jj] == 1)) {
	    p1 = (int)(r1[i].aa - 'A');
	    strncpy(name1, &(atoms->orders[p1][ii * 4]), 4);
	    name1[4] = '\0';

	    p2 = (int)(r2[j].aa - 'A');
	    strncpy(name2, &(atoms->orders[p2][jj * 4]), 4);
	    name2[4] = '\0';

	    if(par->verbose) {
		printf("%s %3d %3d %c %4d %c (%c %4s) %d <=> %3d %3d %c %4d %c (%c %4s) %d = %8.3f ",
		       par->pdb,
		       i, ii, r1[i].pdbnum.cid, r1[i].pdbnum.n, r1[i].pdbnum.in, r1[i].aa, name1, r1[i].t[ii],
		       j, jj, r2[j].pdbnum.cid, r2[j].pdbnum.n, r2[j].pdbnum.in, r2[j].aa, name2, r2[j].t[jj],
		       dist
		       ); 
	    }

	    if(((atoms->use[p1][ii] == 1) && (atoms->use[p2][jj] == 2)) ||
	       ((atoms->use[p1][ii] == 2) && (atoms->use[p2][jj] == 1)) ||
	       ((atoms->use[p1][ii] == 1) && (atoms->use[p2][jj] == 3)) ||
	       ((atoms->use[p1][ii] == 2) && (atoms->use[p2][jj] == 3)) ||
	       ((atoms->use[p1][ii] == 3) && (atoms->use[p2][jj] == 1)) ||
	       ((atoms->use[p1][ii] == 3) && (atoms->use[p2][jj] == 2)) ||
	       ((atoms->use[p1][ii] == 3) && (atoms->use[p2][jj] == 3))) {
		if((dist <= par->hb_tol) && (hb_count1[i] < par->max_hb) && (hb_count2[j] < par->max_hb)) {
		    /* New H-bond */
		    hb_count1[i]++;
		    hb_count2[j]++;
		    contact = 1;
		    if(par->verbose) {
			printf("Hbond ");
		    }
		}
		else if((dist <= par->sb_tol) && (sb_count1[i] < par->max_sb) && (sb_count2[j] < par->max_sb)) {
		    /* New salt-bridge */
		    sb_count1[i]++;
		    sb_count2[j]++;
		    contact = 1;
		    if(par->verbose) {
			printf("Salt-bridge ");
		    }
		} 
	    }
	    else if((r1[i].t[ii] == 0) &&
		    (r2[j].t[jj] == 0) &&
		    (dist <= par->vdw_tol) &&
		    (vdw_count1[i] < par->max_vdw) &&
		    (vdw_count2[j] < par->max_vdw)) {
		/* New VDW contact */
		vdw_count1[i]++;
		vdw_count2[j]++;
		if(par->verbose) {
		    printf("VDW contact");
		}
		contact = 1;
	    }
	    if(par->verbose) {
		printf("\n");
	    }
	}

	if(contact == 1) {
	    nwithin++;
	    if((ii >= 4) && (jj >= 4)) {
		type = 0;
	    }
	    else if(ii < 4) { 
		type = 1; /* First is main chain */
	    }
	    else if(jj < 4) { 
		type = 2; /* Second is main chain */
	    }
	    else {
		type = 3;
	    }
	    cont[i][j][type]++;
	}
    }
    /* printf("Nwithin = %d\n",nwithin);  */

    return cont;
}


int testfile(char *file) {

	FILE *f;
	int flag;

	if((f=fopen(file,"r"))==NULL) {
		flag = 0;
	} else  {
		flag = 1;
		fclose(f);
	}
	return flag;
}





/* Counts the number of domain descriptors in a file ignoring comments */
int count_domain(FILE *IN) {

	int N,col,comment;
	char c;

	N=0;
	comment=0;
	col=0;
	while((c=getc(IN))!=(char)EOF) {
	  if((col==0) && (c=='#' || c=='%')) {
	     comment=1;
	  }
	  if(col!=0 && !comment && c=='{') 
	     N++;
	  col++;
	  if(c=='\n') { 
	     col=0; 
	     comment=0; 
	  }
	}
	return N;
}
int getdomain(FILE *IN, struct domain_loc *domains, int *ndomain, int maxdomain, int *gottrans, char *env, FILE *OUTPUT) {

/* Given a file containing a list of protein descriptors, returns
 *  a list of brookhaven starts and ends, or appropriate wild cards
 *  for subsequent use 
 *
 * New version to remove all the stupid wonk bugs that this used to contain 
 * it now seems very resilient to the placement of newlines, junk, etc */

	int i,j;
	int comment;
	int count,end,nobjects;

	char c;
	char *buff;



	count=0;
	end=0;
	(*gottrans)=0;
	while(!end) {
	  end=domdefine(&domains[count],&i,env,IN,OUTPUT);
/*	  printf("Domain %s\n",domains[count].id);
          printdomain(stdout,domains[count],1); */

	  if(i==1) (*gottrans)=1;
	  if(end==-1) {
	     fprintf(stderr,"error in domain specification file\n");
	     return -1;
	  }
	  count+=(!end);
	  if(count>maxdomain && end!=1) {
	    fprintf(stderr,"error have exceeded maximum domain limit\n");
	    return -1;
	  }
	  if(count==maxdomain) end=1;
	} /* end while(!end)... */
	(*ndomain)=count;

	/* check for duplication */
	for(i=0; i<(*ndomain); ++i) 
	   for(j=i+1; j<(*ndomain); ++j)  
	      if(strcmp(domains[i].id,domains[j].id)==0) {
		 fprintf(stderr,"error: domain identifiers must not be the same\n");
		 fprintf(stderr,"       found two copies of %s, domains %d & %d\n",
			domains[i].id,i+1,j+1);
		 return -1; 
	      }

	return 0;
}

void getdomain_error(char *buff) {
	fprintf(stderr,"error in domain descriptors\n");
	fprintf(stderr,"Last domain read:\n%s\n",buff);
	exit(-1);
}

int domdefine(struct domain_loc *domain, int *gottrans, char *env, FILE *INPUT, FILE *OUTPUT) {

	int i,j,k;
	int nobjects;
	int comment;
	int pt;
	int indom;

 	char c;
	char *descriptor;
	char *buff,*guff;
	char *temp;
	char *dirfile;
	
	FILE *TEST;

	buff=(char*)malloc(2000*sizeof(char));
	guff=(char*)malloc(2000*sizeof(char));
	dirfile=(char*)malloc(500*sizeof(char));
	descriptor=(char*)malloc(2000*sizeof(char));
	(*gottrans)=0;
	
	sprintf(dirfile,"%s/pdb.directories",env);

	/* Now then, the best way to do this is to skip the comments, and
	 *  then just read in a buff starting after the last newline, and
	 *  ending at the end brace */

	indom = 0;
	buff[0]='\0';
	while(fgets(guff,1999,INPUT)!=NULL) {
	     if((guff[0]!='%') && (guff[0]!='#')) {
		if(strstr(guff,"{")!=NULL) { indom=1; }
		if(indom) {
			sprintf(&buff[strlen(buff)],"%s",guff);
		}
		if(strstr(guff,"}")!=NULL) { indom=0; break; }
	     }
	}
		 

	/* First read the file name */

	sscanf(buff,"%s",&domain[0].filename[0]); /* read the filename */
	pt=0;
	if((pt=skiptononspace(buff,pt))==-1) getdomain_error(buff);
	sscanf(&buff[pt],"%s",&domain[0].id[0]); /* read the identifier */
/*	printf("Read in file %s\n",domain[0].filename);  */

	/* check to see whether the file exists, otherwise, look for a file
	 *  with a similar ID */
	if((TEST=fopen(domain[0].filename,"r"))==NULL) {
	  /* look for the file */
	  temp=getfile(domain[0].id,dirfile,4,OUTPUT); /* assume the first four characters are the id */
	  if(temp[0]=='\0') {
	    fprintf(OUTPUT,"file for %s not found, nor was any corrsponding file\n",domain[0].id);
	    fprintf(OUTPUT,"   found in %s\n",dirfile);
	    free(buff); free(dirfile); free(descriptor);
	    free(guff);
	    return -1; 
	  } else {
	    strcpy(&domain[0].filename[0],temp);
	  }
	  free(temp);
	} else {
	  fclose(TEST);
	}
	   

/*	printf("Updated file and id %s and %s\n",domain[0].filename,domain[0].id); 
	printf("Buff is %s\n",buff); 
	printf("Length is %d\n",strlen(buff)); */

	if((pt=skiptononspace(buff,pt))==-1) getdomain_error(buff);
	/* copy the bit between the braces into the string called descriptor */
	i=0; 
	while(buff[pt]!='{' && buff[pt]!='\n' && buff[pt]!='\0') pt++;
	if(buff[pt]=='\n' || buff[pt]=='\0') getdomain_error(buff);
	pt++;
	if(buff[pt]=='\n' || buff[pt]=='\0') getdomain_error(buff);
	j=0; while(buff[pt]!='}' && buff[pt]!='\0') {
	   if(buff[pt]=='\0') getdomain_error(buff);
	   descriptor[j]=buff[pt];
	   pt++; 
	   j++;
	}
	descriptor[j]='\0';
/*	printf("descriptor= '%s'\n",descriptor);     */
	/* allocation of memory, initially */
	domain[0].reverse=(int*)malloc(sizeof(int));
	domain[0].type=(int*)malloc(sizeof(int));
	domain[0].start=(struct brookn*)malloc(sizeof(struct brookn));
	domain[0].end=(struct brookn*)malloc(sizeof(struct brookn)); 
	domain[0].V=(float*)malloc(3*sizeof(float));
	domain[0].v=(float*)malloc(3*sizeof(float));
	domain[0].R=(float**)malloc(3*sizeof(float*));
	domain[0].r=(float**)malloc(3*sizeof(float*));
	for(i=0; i<3; ++i) {
	   domain[0].R[i]=(float*)malloc(3*sizeof(float));
	   domain[0].r[i]=(float*)malloc(3*sizeof(float));
	   for(j=0; j<3; ++j) 
	     if(i==j) domain[0].R[i][j]=domain[0].r[i][j]=1.0;
	     else domain[0].R[i][j]=domain[0].r[i][j]=0.0;
	     domain[0].V[i]=domain[0].v[i]=0.0;
	}

	nobjects=0;
	for(i=0; i<strlen(descriptor); ++i) descriptor[i]=ltou(descriptor[i]);
	pt=0;
	if(strlen(descriptor)==0) getdomain_error(buff);
	while(descriptor[pt]==' ' && descriptor[pt]!='\0') pt++;
	if(descriptor[pt]=='\0' || descriptor[pt]=='}') getdomain_error(buff);
	while(pt!=-1 && descriptor[pt]!='\0' && descriptor[pt]!='\n') { /* read until end of string */
	   if(strncmp(&descriptor[pt],"REVERSE",7)==0) { /* coordinates are to be reversed */
		 domain[0].reverse[nobjects]=1;
		 pt=skiptononspace(descriptor,pt);
	    } else {
		 domain[0].reverse[nobjects]=0;
		 /* don't skip over the text if the word "REVERSE" isn't there */
	   }
	   if(strncmp(&descriptor[pt],"ALL",3)==0) {  /* want all the coordinates in the file */
		 domain[0].type[nobjects]=1;
		 domain[0].start[nobjects].cid=domain[0].start[nobjects].in=
		     domain[0].end[nobjects].cid=domain[0].end[nobjects].in='?';
	 	 domain[0].start[nobjects].n=domain[0].end[nobjects].n=0;
		 pt=skiptononspace(descriptor,pt);
		 nobjects++;
	   } else if(strncmp(&descriptor[pt],"CHAIN",5)==0) { /* want specific chain only */
		 domain[0].type[nobjects]=2;
		 if((pt=skiptononspace(descriptor,pt))==-1) getdomain_error(buff); /* no chain given */
		 domain[0].start[nobjects].cid=domain[0].end[nobjects].cid=descriptor[pt];
		 domain[0].start[nobjects].in=domain[0].end[nobjects].in='?';
		 domain[0].start[nobjects].n=domain[0].end[nobjects].n=0;
		 pt=skiptononspace(descriptor,pt);
		 nobjects++;
	   } else { /* assume that otherwise a specific start and end will be provided */
		 domain[0].type[nobjects]=3;
		 /* cid 1 */
	 	 if(descriptor[pt]=='_') domain[0].start[nobjects].cid=' ';
		 else domain[0].start[nobjects].cid=descriptor[pt];
		 if((pt=skiptononspace(descriptor,pt))==-1) getdomain_error(buff); 
		 /* n 1 */
		 sscanf(&descriptor[pt],"%d",&domain[0].start[nobjects].n);
		 if((pt=skiptononspace(descriptor,pt))==-1) getdomain_error(buff); 
		 /* ins 1 */
		 if(descriptor[pt]=='_') domain[0].start[nobjects].in=' ';
		 else domain[0].start[nobjects].in=descriptor[pt];
		 if((pt=skiptononspace(descriptor,pt))==-1) getdomain_error(buff); 
		 /* skipping over 'to' */
		 if(strncmp(&descriptor[pt],"TO",2)!=0) getdomain_error(buff);
		 if((pt=skiptononspace(descriptor,pt))==-1) getdomain_error(buff); 
		 /* cid 2 */
                 if(descriptor[pt]=='_') domain[0].end[nobjects].cid=' ';
                 else domain[0].end[nobjects].cid=descriptor[pt];
                 if((pt=skiptononspace(descriptor,pt))==-1) getdomain_error(buff); 
                 /* n 2 */
                 sscanf(&descriptor[pt],"%d",&domain[0].end[nobjects].n);
                 if((pt=skiptononspace(descriptor,pt))==-1) getdomain_error(buff); 
                 /* ins 2 */
                 if(descriptor[pt]=='_') domain[0].end[nobjects].in=' ';
                 else domain[0].end[nobjects].in=descriptor[pt];
                 pt=skiptononspace(descriptor,pt);
		 nobjects++;
	   } 
	   if(pt!=-1 && descriptor[pt]=='\n') break;
	   /* reallocing if necessary */
	   if(pt!=-1 && strlen(&descriptor[pt])>0 && descriptor[pt]!='\0' && descriptor[pt]!='\n') {
/*		  printf("Allocating memory for a new object !\n"); */
		  domain[0].reverse=(int*)realloc(domain[0].reverse,(nobjects+1)*sizeof(int));
		  domain[0].type=(int*)realloc(domain[0].type,(nobjects+1)*sizeof(int));
	   	  domain[0].start=(struct brookn*)realloc(domain[0].start,(nobjects+1)*sizeof(struct brookn));
		  domain[0].end=(struct brookn*)realloc(domain[0].end,(nobjects+1)*sizeof(struct brookn));
	   }
	      /* now either stop, or move onto the next descriptor */
        } 
	 
	/* check to see whether there is a transformation */
	if(pt!=-1) { /* there is */
	     while(descriptor[pt]!='\n' && descriptor[pt]!='\0') pt++;
	     if(descriptor[pt]=='\0') getdomain_error(buff);
	     (*gottrans)=1;
	     if(sscanf(&descriptor[pt],"%f%f%f%f%f%f%f%f%f%f%f%f",
		&domain[0].R[0][0],&domain[0].R[0][1],&domain[0].R[0][2],&domain[0].V[0],
		&domain[0].R[1][0],&domain[0].R[1][1],&domain[0].R[1][2],&domain[0].V[1],
		&domain[0].R[2][0],&domain[0].R[2][1],&domain[0].R[2][2],&domain[0].V[2])==(char)EOF)
			 getdomain_error(buff);
/*	     printf("Matrix:\n %f %f %f    %f\n %f %f %f   %f\n %f %f %f   %f\n",
		domain[0].R[0][0],domain[0].R[0][1],domain[0].R[0][2],domain[0].V[0],
                domain[0].R[1][0],domain[0].R[1][1],domain[0].R[1][2],domain[0].V[1],
                domain[0].R[2][0],domain[0].R[2][1],domain[0].R[2][2],domain[0].V[2]);  */
	   }
	   domain[0].nobj=nobjects;
	   free(dirfile);
	   free(descriptor);
	   free(buff);
	   free(guff);
	   return 0;
}

int skiptononspace(char *string, int pointer) {
	while(string[pointer]!=' ' && string[pointer]!='\0' && string[pointer]!='\n') pointer++;
	if(string[pointer]=='\0') return -1;
	while(string[pointer]==' ' && string[pointer]!='\0' && string[pointer]!='\n') pointer++;
	if(string[pointer]=='\0') return -1;
	return pointer;
}

struct astruc *get_clustal(FILE *IN) {

  /* turns X's into '-'s - be warned */
  /* changes sequences to upper case */
   int i,j;
   int w,p,len;
   int first_old;
   char *buff;
   char **t;
   int nt;

   struct astruc *a;

   a = (struct astruc*)malloc(sizeof(struct astruc));
   a->n = 0;
   a->id = (char**)malloc(sizeof(char*));
   a->s  = (char**)malloc(sizeof(char*));

   buff=(char*)malloc(5000*sizeof(char));

   first_old=0;

   while((buff=fgets(buff,4999,IN))!=NULL) {
     if((strncmp(buff,"CLUSTAL",7)==0) || (strncmp(buff,"Clustal",7)==0)) {
/*        printf("%s\n",buff); */
     } else if (strncmp(buff, " ", 1) == 0) {
	 /* ignore the match line */
     } else {
        for(i=0; i<strlen(buff); ++i) { 
           if(buff[i] == '\n') {
             buff[i] = '\0';
             break;
           }
        }
        t = c_split(buff,&nt,' ');
        if((buff[0]!=' ') && (nt == 2)) {
//           printf("nt=%3d %-20s %s\n",nt,t[0],t[1]); 
           /* First see if this is a new or an old identifier */
           w=-1;
           for(i=0; i<a->n; ++i) {
//             printf("Comparing %s to %s\n",a->id[i],t[0]);  
             if(strcmp(a->id[i],t[0])==0) {
                w=i;
                break;
             }
           }
           if(w==-1) {
                w = a->n;
                if(first_old==1) {
                   fprintf(stderr,"WARNING: Odd alignment - new found after old\n");
                }
/*                printf("New id (%d) %s\n",a->n,t[0]); */
                a->id[a->n] = (char*)malloc(MAX_ID_LEN*sizeof(char));
                if((strlen(t[0]))>MAX_ID_LEN) {
                   fprintf(stderr,"MAX_ID_LEN exceeded with %s\n",t[0]);
                   exit(-1);
                }
                strcpy(a->id[a->n],t[0]);
//                printf("Copying %s to %s %d\n",t[0],a->id[a->n],a->n); 
                a->s[a->n]  = (char*)malloc(MAX_SEQ_LEN*sizeof(char));
                a->s[a->n][0]='\0';
                a->n++;
                a->id = (char**)realloc(a->id,((a->n)+1)*sizeof(char*));
                a->s  = (char**)realloc(a->s,((a->n)+1)*sizeof(char*));
/*
 *              printf("Current list\n");
 *              for(i=0; i<a->n; ++i) { 
 *                printf(" -- %s\n",a->id[i]);
 *              }
 */
           } else {
/*                printf("Old id %s\n",t[0]); */
                first_old = 1;
           }
           p = strlen(a->s[w]);
           for(j=0; j<strlen(t[1]); ++j) {
              t[1][j] = ltou(t[1][j]);
           }
//           printf("String goes from:\n this %s to \n",a->s[w]);
           strcpy(&(a->s[w][p]),t[1]);
//           printf(" this %s\n",a->s[w]);
           
        }
        for(i=0; i<nt; ++i) { free(t[i]); }
        free(t);
     }
   }
/*   printf("Found %d entries\n",a->n); *a

   /* Some Checks */
   if((a->n)>1) {
     len = strlen(a->s[0]);
     for(i=1; i<a->n; ++i) {
        if(strlen(a->s[i]) != len) {
           fprintf(stderr,"get_clustal: length mis-match %s = %d vs %s = %d\n",
              a->id[0],len,a->id[i],strlen(a->s[i]));
           fprintf(stderr,"Strings are:\n%s %4d %s\n%s %4d %s\n", a->id[0],len,a->s[0],a->id[i],strlen(a->s[i]),a->s[i]);
           exit(-1);
        }
        /* change 'X' to '-' */
        for(j=0; j<(strlen(a->s[i])); ++j) {
           a->s[i][j] = ltou(a->s[i][j]);
           if(a->s[i][j] == 'X') {
            a->s[i][j] = '-';
           }
        }
     }
   }
   a->len = strlen(a->s[0]);
   free(buff);
   return a;
}

int *map_coord_align(struct astruc *a, char *id, struct residue *r, int nr, struct parameters *par) {
    int i,j;
    int w;
    int *p;
    char aa;

    if(par->verbose) {
      printf("#Alen = %d\n",a->len);
      printf("#NR   = %d\n",nr);
    }
    p = (int*)malloc(nr*sizeof(int));
    w=-1;
    for(i=0; i<a->n; ++i) {
      if(strcmp(a->id[i],id)==0) {
        w=i;
        break;
      }
    }
    if(w==-1) {
      fprintf(stderr,"ERROR: id %s not found in alignment\n",id);
      exit(-1);
    }
    j=0;  /* position in alignment */
    for(i=0; i<nr; ++i) {
      aa = a->s[w][j];
      if(par->really_verbose) {
        printf("Trying %d %c\n",i,r[i].aa);
      }
      while(((aa == '-') || (aa == '.') || (aa == ' ') || (aa == '_')) && (j<(a->len))) {
          j++;
          aa = a->s[w][j];
      }
      if((j>=(a->len)) || (aa != r[i].aa)) {
        fprintf(stderr,"MAP_COORD_ALIGN: ID %s mismatch between alignment & structure at structure pos %d (%c vs %c)\n",
            id,i,aa,r[i].aa);
        exit(-1);
      }
      if(par->really_verbose) {
        printf("Equating %d %c with %d %c\n",i,r[i].aa,j,aa);
      }
      p[i] = j;
      j++;
      if(j>=(a->len)) { break; }
    }
    return p;
}

void get_pp_scores(char aa_i, char aa_j, int ss, int sm, int ms, float *val_ss, float *val_sm, float *val_ms, struct pp_mat *pp, struct parameters *par) {

    int p_i, p_j;

    if((aa_i >= 'A') && (aa_i <='Z')) {
	p_i = (int)(aa_i - 'A');
    } 
    else {
	p_i = -1;
    }

    if((aa_j >= 'A') && (aa_j <='Z')) {
	p_j = (int)(aa_j - 'A');
    } 
    else {
	p_j = -1;
    }

    if((p_i == -1) || (p_j == -1)) {
	*val_ss = -1.0;
    } 
    else {
	*val_ss = pp->p[p_i][p_j];
    }

    if(ss < par->nc_min) { // Only if cont made
       *val_ss = 0.0;
    }

    // Displaying the side-chain to main-chain value
    *val_ms = 0.0;
    if(sm >= par->nc_min) {
	if((p_i == -1) || (p_j == -1)) {
	    *val_ms = -0.5;
	}
	else {
	    *val_ms = pp->m[p_j];
	}
    }

    *val_sm = 0.0; /* changed from 'val_ms = 0.0' - a typo... that resulted in val_sm unitialised... */
    if(ms > par->nc_min) { /* Main-chain in second */
	if((p_i == -1) || (p_j == -1)) {
	    *val_sm = -0.5;
	}
	else {
	    *val_sm = pp->m[p_i];
	}
    }
}

float int_score(struct astruc *a1, int *p1, struct residue *r1, int nr1, int pred1, 
                struct astruc *a2, int *p2, struct residue *r2, int nr2, int pred2, 
                int ***cont, struct pp_mat *pp,  int ras, struct parameters *par) {

     int i,j;
     int p_i, p_j;
     int wval;

     char aa_i, aa_j;
     float val;
     float score;
  
     FILE *RAS;

     if(ras==1) {
       if((RAS=fopen(par->ras_file,"w"))==NULL) {
         fprintf(stderr,"Error writing %s\n",par->ras_file);
         exit(-1);
       }
       fprintf(RAS,"zap\n");
       fprintf(RAS,"load pdb%s.ent\n",par->pdb);
       fprintf(RAS,"wireframe off\n");
       fprintf(RAS,"colour chain\n");
       fprintf(RAS,"backbone 10\n");
     }
       
     score=0.0; 
     for(i=0; i<nr1; ++i) {
      aa_i = ltou(a1->s[pred1][p1[i]]);
      if((aa_i >= 'A') && (aa_i <='Z')) {
        p_i = (int)(aa_i - 'A');
      } else {
        p_i = -1;
      }
      for(j=0; j<nr2; ++j) {
         aa_j = ltou(a2->s[pred2][p2[j]]);
         if((aa_j >= 'A') && (aa_j <='Z')) {
           p_j = (int)(aa_j - 'A');
         } else {
           p_j = -1;
         }
         if(cont[i][j][0]>0) { /* Side-side */
           if((p_i == -1) || (p_j == -1)) {
               val = -1.0;  
           } else {
               val = pp->p[p_i][p_j];
           }
           score += val;
           if(par->verbose) {
             printf(" ORIG SS: cont %3d %c %4d %c <-> %3d %c %4d %c : %c <-> %c Mutates to %c <-> %c = %6.2f\n",
                 i,r1[i].pdbnum.cid,r1[i].pdbnum.n,r1[i].pdbnum.in,
                 j,r2[j].pdbnum.cid,r2[j].pdbnum.n,r2[j].pdbnum.in,
                 r1[i].aa, r2[j].aa,aa_i,aa_j,val);
           }
           if(ras==1) {
             if(val>1.0) { wval = 200; }
             else if(val>0.0) { wval = 100; }
             else { wval = 10; }
             fprintf(RAS,"select ((*%c AND %d) OR (*%c and %d)) AND (sidechain OR *.CA)\n",
                r1[i].pdbnum.cid,r1[i].pdbnum.n, r2[j].pdbnum.cid,r2[j].pdbnum.n);
             fprintf(RAS,"wireframe %d\n",wval);
             fprintf(RAS,"colour shapely\n");
           }
          } 
          if(cont[i][j][1]>0) { /* Main-chain in first */
              if((p_i == -1) || (p_j == -1)) {
                 val = -0.5;
              } else {
                 val = pp->m[p_j];
              }
              score += val;
              if(par->verbose) {
                printf(" ORIG MS: cont %3d %c %4d %c <-> %3d %c %4d %c : %c <-> %c Mutates to %c <-> %c = %6.2f\n",
                 i,r1[i].pdbnum.cid,r1[i].pdbnum.n,r1[i].pdbnum.in,
                 j,r2[j].pdbnum.cid,r2[j].pdbnum.n,r2[j].pdbnum.in,
                 r1[i].aa, r2[j].aa,aa_i,aa_j,val);
              }
              if(ras==1) {
                if(val>1.0) { wval = 200; }
                else if(val>0.0) { wval = 100; }
                else { wval = 10; }
                fprintf(RAS,"select (*%c AND %d AND mainchain) OR (*%c and %d AND (sidechain OR *.CA))\n",
                   r1[i].pdbnum.cid,r1[i].pdbnum.n, r2[j].pdbnum.cid,r2[j].pdbnum.n);
                fprintf(RAS,"wireframe %d\n",wval);
                fprintf(RAS,"colour shapely\n");
              }
           } 
           if(cont[i][j][2]>0) { /* Main-chain in second */
              if((p_i == -1) || (p_j == -1)) {
                val = -0.5;
              } else {
                val = pp->m[p_i];
              }
              score += val;
              if(par->verbose) {
                printf(" ORIG SM: cont %3d %c %4d %c <-> %3d %c %4d %c : %c <-> %c Mutates to %c <-> %c = %6.2f\n",
                 i,r1[i].pdbnum.cid,r1[i].pdbnum.n,r1[i].pdbnum.in,
                 j,r2[j].pdbnum.cid,r2[j].pdbnum.n,r2[j].pdbnum.in,
                 r1[i].aa, r2[j].aa,aa_i,aa_j,val);
              }
              if(ras==1) {
                if(val>1.0) { wval = 200; }
                else if(val>0.0) { wval = 100; }
                else { wval = 10; }
                fprintf(RAS,"select (*%c AND %d AND (sidechain OR *.CA)) OR (*%c and %d AND mainchain)\n",
                   r1[i].pdbnum.cid,r1[i].pdbnum.n, r2[j].pdbnum.cid,r2[j].pdbnum.n);
                fprintf(RAS,"wireframe %d\n",wval);
                fprintf(RAS,"colour shapely\n");
              }
            }
         }
      }
      if(ras==1) {
       fclose(RAS);
      }
      return score;
}
int main(int argc, char **argv) {
   
   int i,j,k,l,m,n,o,p,q,r;
   int f,c;
   int best, which;
   int ndomain,trans;
   int nr1,nr2;
   int na1,na2;
   int ***cont;
   int *p1,*p2;
   int *change1,*change2;
   int w1,w2;
   int p_k, p_l;
   int ntotal,nbetter;
   int np,ras;
   int n_ss,n_sm,n_mm;
   int aa_i,aa_j,p_i,p_j;

   float val,val_ss,val_sm,val_ms;
   float rf,rv;
   float sum,sumsq;
   float ave,sd,Z;
   float **score;
   float rscore;
   float ref_score,ref_Z;
   float pvalue,min_pvalue;
   float fc1,fc2;
   float *fchange1,*fchange2;

   char *env;
   char *orig1, *orig2;
   char buff[1000];
   char name[1000];
   char os[1000];
   char outfile[1000];
   char AARAND[2000];
   char pdb[5];
   char aa_k, aa_l;

   struct parameters *par;
   struct dist ***distmat;
   struct astruc *a1;
   struct astruc *a2;
   struct residue *r1;
   struct residue *r2;
   struct atom_details *atoms;
   struct domain_loc *domain;
   struct pp_mat *pp;

   FILE *AFILE, *DFILE, *OUT, *ATOMS;
   FILE *SS, *SM;

   if(argc<2) exit_error();

   par=(struct parameters*)malloc(sizeof(struct parameters));

   /* get environment variable */
   if((env=getenv("I2DIR"))==NULL) {
     fprintf(stderr,"error: environment variable I2DIR must be set\n");
     exit(-1);
   }
   par->mode           =      1; /* Run interprets in default mode */
   par->quiet          =      0; /* Quiet mode */
   par->disp           =      0; /* Output rasmol script */
   par->verbose        =      0;  
   par->really_verbose =      0;
   par->ca_ca_tol      =   15.0;
   par->hb_tol         =    3.5;
   par->max_hb         =    100;
   par->sb_tol         =    5.5;
   par->max_sb         =    100;
   par->vdw_tol        =    4.5;
   par->max_vdw        =    100;
   par->nrand          =      0;
   par->rand           =      1;
   par->diff           =      0;
   par->nc_min         =      1; /* number of contacts to be seen within d_tol to count */
   par->show_template  =      0; /* don't show template by default */
   par->id_min         =   30.0; /* range down to mutate (mode=4) */
   strcpy(par->dom_file,"default.dom");
   strcpy(par->ras_file,"i2_ras.script");
   strcpy(par->aln_file1,"default1.aln");
   strcpy(par->aln_file2,"default2.aln");
   outfile[0]='\0';
   sprintf(par->ats_file,"%s/atoms.dat",env);
   sprintf(par->ss_mat_file,"%s/ss_mat.dat",env);
   sprintf(par->sm_mat_file,"%s/sm_mat.dat",env);

   for(i=1; i<argc; ++i) {
      if(argv[i][0]!='-') exit_error();
      if(strcmp(&argv[i][1],"d")==0) {
        if((i+1)>=argc) exit_error();
        strcpy(&par->dom_file[0],argv[i+1]);
        i++;
      } else if((strcmp(&argv[i][1],"a")==0)) {
        if((i+2)>=argc) exit_error();
        strcpy(&par->aln_file1[0],argv[i+1]);
        strcpy(&par->aln_file2[0],argv[i+2]);
        i+=2;
      } else if(strcmp(&argv[i][1],"ras")==0) {
        if((i+1)>=argc) exit_error();
        strcpy(&par->ras_file[0],argv[i+1]);
        i++;
      } else if((strcmp(&argv[i][1],"op")==0)) {
        if((i+2)>=argc) exit_error();
        strcpy(&par->p1[0],argv[i+1]);
        strcpy(&par->p2[0],argv[i+2]);
        par->onepair=1;
        i+=2;
      } else if(strcmp(&argv[i][1],"ss_mat")==0) {
        if((i+1)>=argc) exit_error();
      } else if(strcmp(&argv[i][1],"ss_mat")==0) {
        if((i+1)>=argc) exit_error();
        strcpy(par->ss_mat_file,argv[i+1]);
        i++;
      } else if(strcmp(&argv[i][1],"sm_mat")==0) {
        if((i+1)>=argc) exit_error();
        strcpy(par->sm_mat_file,argv[i+1]);
        i++;
      } else if((strcmp(&argv[i][1],"ca_ca_tol")==0)) {
        if((i+1)>=argc) exit_error();
        sscanf(argv[i+1],"%f",&par->ca_ca_tol);
        i++;
      } else if((strcmp(&argv[i][1],"diff")==0)) {
        par->diff=1;
      } else if((strcmp(&argv[i][1],"hb_tol")==0)) {
        if((i+1)>=argc) exit_error();
        sscanf(argv[i+1],"%f",&par->hb_tol);
        i++;
      } else if((strcmp(&argv[i][1],"rand")==0)) {
        if((i+1)>=argc) exit_error();
        sscanf(argv[i+1],"%d",&par->nrand);
        i++;
      } else if(((strcmp(&argv[i][1],"randmode")==0)) || ((strcmp(&argv[i][1],"mode")==0))) {
        if((i+1)>=argc) exit_error();
        sscanf(argv[i+1],"%d",&par->rand);
        if((par->rand<1) || (par->rand>4)) { exit_error(); }
        i++;
      } else if((strcmp(&argv[i][1],"show_template")==0)) {
        par->show_template=1;
      } else if((strcmp(&argv[i][1],"sb_tol")==0)) {
        if((i+1)>=argc) exit_error();
        sscanf(argv[i+1],"%f",&par->sb_tol);
        i++;
      } else if((strcmp(&argv[i][1],"vdw_tol")==0)) {
        if((i+1)>=argc) exit_error();
        sscanf(argv[i+1],"%f",&par->vdw_tol);
        i++;
      } else if((strcmp(&argv[i][1],"nc_min")==0)) {
        if((i+1)>=argc) exit_error();
        sscanf(argv[i+1],"%d",&par->nc_min);
        i++;
      } else if((strcmp(&argv[i][1],"fasta")==0)) {
        par->mode=2;
        par->quiet=1;
      } else if((strcmp(&argv[i][1],"count")==0)) {
        par->mode=3;
        par->quiet=1;
      } else if((strcmp(&argv[i][1],"q")==0)) {
        par->quiet=1;
      } else if((strcmp(&argv[i][1],"disp")==0)) {
        par->disp=1;
      } else if((strcmp(&argv[i][1],"help")==0) || (strcmp(&argv[i][1],"h")==0)) { 
        i2_help(par);
      } else if(strcmp(&argv[i][1],"v")==0 || strcmp(&argv[i][1],"V")==0) {
        par->verbose=1;
      } else if(strcmp(&argv[i][1],"rv")==0 || strcmp(&argv[i][1],"RV")==0) { par->really_verbose=1; } else {
        exit_error();
      }
   }
   if(par->nrand==0) { par->rand=0; }

   if(strlen(outfile)>=1) {
     if((OUT=fopen(outfile,"w"))==NULL) {
       fprintf(OUT,"error opening file %s\n",outfile);
       exit(-1);
     }
   } else {
     OUT = stdout;
   }


   if(par->quiet==0) {
     fprintf(OUT,"# InterPReTS 2\n");
     fprintf(OUT,"# Developed by:\n");
     fprintf(OUT,"#  Patrick Aloy & Rob Russell\n");
     fprintf(OUT,"#  EMBL, Meyerhofstrasse 1, 69117 Heidelberg, Germany\n");
     fprintf(OUT,"#  Contact: russell@embl.de\n");
     fprintf(OUT,"# When using, please cite:\n");
     fprintf(OUT,"#  P. Aloy & R.B. Russell\n");
     fprintf(OUT,"#  Interrogating protein interaction networks through structural biology.\n");
     fprintf(OUT,"#  Proc. Natl. Acad. Sci. USA, 99, 5896-5901, 2002 (PubMed Id 11972061).\n");
     fprintf(OUT,"#\n");
   }

   if((ATOMS=fopen(par->ats_file,"r"))==NULL) {
         printf("error opening file %s\n",par->ats_file);
         exit(-1);
   }
   atoms = get_atom_details(ATOMS);
   close(ATOMS);

   if(par->mode==1) { /* read pair-potential matrix */
     if((SS=fopen(par->ss_mat_file,"r"))==NULL) {
         printf("error opening file %s\n",par->ss_mat_file);
         exit(-1);
     }
     if((SM=fopen(par->sm_mat_file,"r"))==NULL) {
         printf("error opening file %s\n",par->sm_mat_file);
         exit(-1);
     }
     pp = read_pp_mat(SS,SM,par);
     fclose(SS);
     fclose(SM);
   }

   if(par->mode==1) {
     if((AFILE=fopen(par->aln_file1,"r"))==NULL) {
       fprintf(OUT,"error opening file %s\n",par->aln_file1);
       exit(-1);
     }
     a1 = get_clustal(AFILE);
     fclose(AFILE);
     if(par->quiet==0) {
       printf("# File %-10s %3d sequences, length %d\n",par->aln_file1,a1->n,strlen(a1->s[0]));
     }

     if((AFILE=fopen(par->aln_file2,"r"))==NULL) {
       fprintf(OUT,"error opening file %s\n",par->aln_file2);
       exit(-1);
     }
     a2 = get_clustal(AFILE);
     fclose(AFILE);  

     if(par->quiet==0) {
       printf("# File %-10s %3d sequences, length %d\n",par->aln_file2,a2->n,strlen(a2->s[0]));
     }
   }

   
   if((DFILE=fopen(par->dom_file,"r"))==NULL) {
     fprintf(OUT,"error opening file %s\n",par->dom_file);
     exit(-1);
   }
   ndomain=count_domain(DFILE);
   domain=(struct domain_loc*)malloc(ndomain*sizeof(struct domain_loc));
   rewind(DFILE);

   if(getdomain(DFILE,domain,&ndomain,ndomain,&trans,env,OUT)==-1) {
     exit(-1);
   }
   fclose(DFILE);
   if(par->quiet==0) {
      printf("# File %-10s %3d domains\n",par->dom_file,ndomain);
   }
   r1 = getallcoords(&domain[0], &nr1, &na1, domain[0].filename, atoms->orders, par->verbose, par->really_verbose);
   domain[0].ncoords = nr1;
   if(par->quiet==0) {
      printf("# Domain %-20s %4d residues %7d atoms\n", domain[0].id,nr1,na1);
   }
   r2 = getallcoords(&domain[1], &nr2, &na2, domain[1].filename, atoms->orders, par->verbose, par->really_verbose);
   if(par->quiet==0) {
     printf("# Domain %-20s %4d residues %7d atoms\n", domain[1].id,nr2,na2);
   }
   domain[1].ncoords = nr2;

   if(par->mode==2) {
     /* Just output FASTA and die */
     printf(">%s\n",domain[0].id);
     for(j=0; j<nr1; ++j) {
       printf("%c",r1[j].aa);
       if(((j+1)%60)==0) {  printf("\n"); }
     }
     printf("\n");
     printf(">%s\n",domain[1].id);
     for(j=0; j<nr2; ++j) {
       printf("%c",r2[j].aa);
       if(((j+1)%60)==0) {  printf("\n"); }
     }
     printf("\n");
     exit(0);
   }

   if(par->mode==1) {
     /* Find out which entries in the alignment are structures */
     w1=-1;
     for(i=0; i<a1->n; ++i) {
       if(strcmp(a1->id[i],domain[0].id)==0) {
         w1 = i;
         break;
       }
     }
     if(w1 == -1) {
       fprintf(stderr,"Error finding %s in alignment %s - ids must match exactly\n",
         domain[0].id,par->aln_file1);
       exit(-1);
     }
     w2=-1;
     for(i=0; i<a2->n; ++i) {
       if(strcmp(a2->id[i],domain[1].id)==0) {
         w2 = i;
         break;
       }
     }
     if(w2 == -1) {
       fprintf(stderr,"Error finding %s in alignment %s - ids must match exactly\n",
         domain[1].id,par->aln_file2);
       exit(-1);
     }
     /* Map coordinates to alignment */
     p1 = map_coord_align(a1, domain[0].id, r1, nr1, par);
     p2 = map_coord_align(a2, domain[1].id, r2, nr2, par);
   }

   /* Get contacts */
   if(par->mode == 3) {
       par->verbose = 1;
   }
   strncpy(par->pdb, domain[0].id, 4);
   par->pdb[4] = '\0';
   cont = getcontacts(r1, nr1, r2, nr2, atoms, par);
   if(par->mode == 3) {
       exit(-1);
   }

   if(par->mode == 1) {
     /* Set scores to zero */
     score = (float **) malloc(a1->n * sizeof(float *));
     for(k = 0; k < a1->n; ++k) {
	 score[k] = (float*)malloc(a1->n*sizeof(float));
	 for(l = 0; l < a2->n; ++l) {
	     score[k][l] = 0.0;
	 }
     }
   }

   
   /* Now compute the scores */
   n_ss = 0;
   n_sm = 0;
   n_mm = 0;
   for(i = 0; i < nr1; ++i) {
       for(j = 0; j < nr2; ++j) {
	   os[0] = '\0';
	   if(cont[i][j][0] >= par->nc_min) {
	       sprintf(os,"SS %2d ", cont[i][j][0]);
	       n_ss++;
	   }
	   if(cont[i][j][1] >= par->nc_min) {
	       sprintf(&os[strlen(os)], "MS %2d ", cont[i][j][1]);
	       n_sm++;
	   }
	   if(cont[i][j][2] >= par->nc_min) {
	       sprintf(&os[strlen(os)], "SM %2d ", cont[i][j][2]);
	       n_sm++;
	   }
	   if(cont[i][j][3] > par->nc_min) {
	       sprintf(&os[strlen(os)], "MM %2d ", cont[i][j][3]);
	       n_mm++;
	   }
	   if((os[0]!='\0') && (par->quiet==0)) {
	       // Displaying the side-chain to side-chain value
	       aa_i = ltou(r1[i].aa);
	       aa_j = ltou(r2[j].aa);

	       get_pp_scores(aa_i, aa_j, cont[i][j][0], cont[i][j][1], cont[i][j][2], &val_ss, &val_sm, &val_ms, pp, par);

	       printf("#%-20s cont %3d %c %4d %c <-> %3d %c %4d %c : %c <-> %c %5.2f %5.2f %5.2f",
		      os,
		      i, r1[i].pdbnum.cid, r1[i].pdbnum.n, r1[i].pdbnum.in,
		      j, r2[j].pdbnum.cid, r2[j].pdbnum.n, r2[j].pdbnum.in,
		      r1[i].aa, r2[j].aa, val_ss, val_sm, val_ms
		      );

	       if(par->mode == 1) {
		   printf(" : ");
		   for(k=0; k < a1->n; ++k) { 
		       if(k == w1) {
			   printf("[");
		       }
		       printf("%c", a1->s[k][p1[i]]); 
		       if(k == w1) {
			   printf("]");
		       }
		   }
		   printf(" <-> ");
		   for(k = 0; k < a2->n; ++k) { 
		       if(k == w2) {
			   printf("[");
		       }
		       printf("%c", a2->s[k][p2[j]]); 
		       if(k == w2) {
			   printf("]");
		       }
		   }
	       }
	       printf("\n");
	   }
       }
   }

   printf("# N sidechain-sidechain %4d N sidechain-mainchain %4d N mainchain-mainchain %4d\n",n_ss,n_sm,n_mm);
   if(par->mode==1) { /* computing scores */
    if(par->nrand>0) { /* Do random shuffles */
     srand(time(NULL));
     p=0;
     for(i=0; i<strlen(AA1a); ++i) {
       for(j=0; j<(int)(10*AA_ABUN[i]); ++j) {
          AARAND[p]=AA1a[i];
          p++;
       }
     }
     AARAND[p] = '\0';
/*     printf("AARAND %s\n",AARAND);  */
     orig1 = (char*)malloc(a1->len*sizeof(char));
     orig2 = (char*)malloc(a2->len*sizeof(char));

     if(par->rand==1) { /* Just on the template */
        strcpy(orig1,a1->s[w1]);
        strcpy(orig2,a2->s[w2]);
        sumsq = 0.0;
        sum   = 0.0;
        n     = 0;
        for(i=0; i<par->nrand; ++i) {
         for(j=0; j<a1->len; ++j) {
             if((a1->s[w1][j]>='A') && ((a1->s[w1][j]<='Z'))) {
               rf = (float)(rand()/(float)RAND_MAX);
               r = (int)(rf*strlen(AARAND));
               a1->s[w1][j] = AARAND[r];
/*               printf("%6.3f %3d %c\n",rf,r,AARAND[r]); */
             }
         }
         p = par->verbose; 
         par->verbose=0;
         rscore =  int_score(a1,p1,r1,nr1,w1,a2,p2,r2,nr2,w2,cont,pp,0,par);
         par->verbose=p;
         sum += rscore;
         sumsq += rscore*rscore;
         n++;
       }
       strcpy(a1->s[w1],orig1);
       for(i=0; i<par->nrand; ++i) {
         for(j=0; j<a2->len; ++j) {
               if((a2->s[w2][j]>='A') && ((a2->s[w2][j]<='Z'))) {
                 rf = (float)(rand()/(float)RAND_MAX);
                 r = (int)(rf*strlen(AARAND));
                 a2->s[w2][j] = AARAND[r];
/*                 printf("%6.3f %3d %c\n",rf,r,AARAND[r]); */
               }
          }
          p = par->verbose; par->verbose=0;
          rscore =  int_score(a1,p1,r1,nr1,w1,a2,p2,r2,nr2,w2,cont,pp,0,par);
          par->verbose=p;
          sum += rscore;
          sumsq += rscore*rscore;
          n++;
       }
       strcpy(a2->s[w2],orig2);
       ave = (float)sum/(float)n;
       sd =  sqrt( (sumsq - ((sum * sum)/(float)n))/((float)(n-1)));
      }
     }
     ref_score = 0.0;
     for(k=0; k<a1->n; ++k) {
      for(l=0; l<a2->n; ++l) {
        if((par->onepair==0) || 
           ((strcmp(a1->id[k],par->p1)==0) && (strcmp(a2->id[l],par->p2)==0)) ) {
         if(par->verbose) {
            printf("TESTING: %s <=> %s fit on %s <=> %s\n", 
                a1->id[k],a2->id[l],a1->id[w1],a2->id[w2]);
         }
         if((np==1) && (par->disp==1)) {
            ras=1;
         }
         score[k][l] = int_score(a1,p1,r1,nr1,k,a2,p2,r2,nr2,l,cont,pp,ras,par);
         if((k==w1) && (l==w2) && ((par->rand==3) || (par->rand==4))) { 
          ref_score = score[k][l];
         }
       }
      }
     }

     /* Computing differences from each to template */
     change1=(int*)malloc((a1->n)*sizeof(int));
     fchange1=(float*)malloc((a1->n)*sizeof(float));
     for(k=0; k<a1->n; ++k) {
       change1[k]=0;
       n=0;
       for(l=0; l<a1->len; ++l) {
         if((a1->s[k][l]>='A') && ((a1->s[k][l]<='Z'))) {
            n++;
            if((a1->s[k][l]!=a1->s[w1][l])) {
              change1[k]++;
            }
         }
       }
       fchange1[k] = (float)change1[k]/(float)n;
//       printf("%-20s %4d %8.3f\n",a1->id[k],change1[k],fchange1[k]);
     }
     fchange2=(float*)malloc((a2->n)*sizeof(float));
     change2=(int*)malloc((a2->n)*sizeof(int));
     for(k=0; k<a2->n; ++k) {
       change2[k]=0;
       n=0;
       for(l=0; l<a2->len; ++l) {
         if((a2->s[k][l]>='A') && ((a2->s[k][l]<='Z'))) {
            n++;
            if(a2->s[k][l]!=a2->s[w2][l]) {
              change2[k]++;
            }
         }
       }
       fchange2[k] = (float)change2[k]/(float)n;
//       printf("%-20s %4d %8.3f\n",a2->id[k],change2[k],fchange2[k]);
     }
     printf("%%Fields below:\n");

     if(par->show_template==1) {
       printf("%% id template A, id template B\n");
     }
     printf("%% id A', %% sequence id A-A'\n");
     printf("%% id B', %% sequence id B-B'\n");
     if((par->rand==3) || (par->rand==4)) {
       printf("%% Raw score (interprets score for model - that for template)\n");
     } else {
       printf("%% Raw score (interprets score for model)\n");
     }
     if(par->rand!=0) {
       printf("%% randomisations mean, sd, Z, p [randmode 2 or 3 only], orig-score [mode 3 only]\n");
       printf("%%  (incidentally, randmode is %d)\n",par->rand);
     }
   
     np=0; 
     for(k=0; k<a1->n; ++k) {
       for(l=0; l<a2->n; ++l) {
        if((par->onepair==0) || 
           ((strcmp(a1->id[k],par->p1)==0) && (strcmp(a2->id[l],par->p2)==0))) {

           np++;

           if(par->diff==1) {
             score[k][l] -= ref_score;
           }
           printf("SUM");
           if((k==w1) && (l==w2)) { 
             printf("* "); 
           } else { 
             printf("  "); 
           }
           if(par->show_template==1) {
             printf("%-20s %-20s ",a1->id[w1],a2->id[w2]);
           }
           printf("%-20s %6.2f %-20s %6.2f %8.3f ",
                a1->id[k],(100.00*(1.0-fchange1[k])),
                a2->id[l],(100.00*(1.0-fchange2[l])),
                score[k][l]);

           /* Randomisations */
           if(par->nrand>0) { 
            if(par->rand==2) {
             nbetter=0;
             ntotal=0;
             strcpy(orig1,a1->s[k]);
             strcpy(orig2,a2->s[l]);
             sumsq = 0.0;
             sum   = 0.0;
             n     = 0;
             for(i=0; i<par->nrand; ++i) {
                for(j=0; j<a1->len; ++j) {
                   /* rand==2 is akin to the interprets thing, but run for each sequence - 
                    * that is randomising the query sequences, not the template
                    * rand==3 is the new way - mutating the template according to the 
                    * rand==4 is like 3, but mutating everything down to a fixed level par->id_min
                    * similarity to the query sequences */

                   if((a1->s[k][j]>='A') && ((a1->s[k][j]<='Z'))) {
                     rf = (float)(rand()/(float)RAND_MAX);
                     r = (int)(rf*strlen(AARAND));
                     a1->s[k][j] = AARAND[r];
    /*                 printf("%6.3f %3d %c\n",rf,r,AARAND[r]); */
                   }
                }
    /*            printf("\nORIG:%s\nRAND:%s\n",orig1,a1->s[k]); */
                p = par->verbose; par->verbose=0;
                rscore =  int_score(a1,p1,r1,nr1,k,a2,p2,r2,nr2,l,cont,pp,0,par);
                if(par->diff==1) {
                  rscore -= ref_score;
                }
                par->verbose=p;
    /*            printf("Random %3d %8.3f\n",n,rscore); */
                sum += rscore;
                sumsq += rscore*rscore;
                if(rscore>=score[k][l]) {
                    nbetter++;
                } 
                ntotal++;
                n++;
             }
             strcpy(a1->s[k],orig1);
             for(i=0; i<par->nrand; ++i) {
                for(j=0; j<a2->len; ++j) {
                   if((a2->s[l][j]>='A') && ((a2->s[l][j]<='Z'))) {
                     rf = (float)(rand()/(float)RAND_MAX);
                     r = (int)(rf*strlen(AARAND));
                     a2->s[l][j] = AARAND[r];
    /*                 printf("%6.3f %3d %c\n",rf,r,AARAND[r]); */
                   }
                }
    /*            printf("\nORIG:%s\nRAND:%s\n",orig2,a2->s[l]); */
                rscore =  int_score(a1,p1,r1,nr1,k,a2,p2,r2,nr2,l,cont,pp,0,par);
                if(par->diff==1) {
                  rscore -= ref_score;
                }
    /*            printf("Random %3d %8.3f\n",n,rscore); */
                sum += rscore;
                sumsq += rscore*rscore;
                if(rscore>=score[k][l]) {
                    nbetter++;
                } 
                ntotal++;
                n++;
             }
             strcpy(a2->s[l],orig2);
             ave = (float)sum/(float)n;
             sd =  sqrt( (sumsq - ((sum * sum)/(float)n))/((float)(n-1)));
             /* spec_rand_sd=sqrt((sumsq2-(sum2*sum2)/(float)NRAND2) /((float)(NRAND2-1))); */
            } else if((par->rand==3) || (par->rand==4)) {
    //         printf("Here\n");
             strcpy(orig1,a1->s[w1]);
             strcpy(orig2,a2->s[w2]);
             sumsq = 0.0;
             sum   = 0.0;
             n     = 0;
             nbetter = 0;
             ntotal  = 0;
             if(par->rand==4) {
                 fc1 = (float)(100-par->id_min)/100.00;
                 fc2 = (float)(100-par->id_min)/100.00;
             } else {
                 fc1 = fchange1[k];
                 fc2 = fchange2[l];
             }
             for(i=0; i<par->nrand; ++i) {
                for(j=0; j<a1->len; ++j) {
                   /* rand==3 is the new way - mutating the template according to the 
                    * similarity to the query sequences 
                    * we only change the position roughly according to the seq id.  
                    * This is done by testing whether a random value is less than or equal to the 
                    * seq sim (0-1) */
                 if((a1->s[w1][j]>='A') && ((a1->s[w1][j]<='Z'))) {
                   rv = (float)rand()/(float)RAND_MAX;
                   if(rv<fc1) {
    //                 printf("Res %c ",a1->s[w1][j]);
                     rf = (float)(rand()/(float)RAND_MAX);
                     r = (int)(rf*strlen(AARAND));
                     a1->s[w1][j] = AARAND[r];
    //                 printf("%6.3f %3d %c\n",rf,r,AARAND[r]); 
    //                 printf("%c %8.3f compared to %8.3f\n",a1->s[w1][j],rv,fchange1[k]);
                   }
                 }
                }
                for(j=0; j<a2->len; ++j) {
                  if((a2->s[w2][j]>='A') && ((a2->s[w2][j]<='Z'))) { 
                   rv = (float)rand()/(float)RAND_MAX;
                   if(rv<fc2) {
    //                 printf("Res %c ",a2->s[w2][j]);
                     rf = (float)(rand()/(float)RAND_MAX);
                     r = (int)(rf*strlen(AARAND));
                     a2->s[w2][j] = AARAND[r];
    //                 printf("%6.3f %3d %c\n",rf,r,AARAND[r]); 
    //                 printf("%c %8.3f compared to %8.3f\n",a2->s[w2][j],rv,fchange2[l]);
                   }
                  }
                }
    /*            printf("\nORIG:%s\nRAND:%s\n",orig1,a1->s[k]); */
                p = par->verbose; par->verbose=0;
                rscore =  int_score(a1,p1,r1,nr1,w1,a2,p2,r2,nr2,w2,cont,pp,0,par);
                if(par->diff==1) {
                  rscore -= ref_score;
               }
                par->verbose=p;
    /*            printf("Random %3d %8.3f\n",n,rscore); */
                sum += rscore;
                sumsq += rscore*rscore;
                if(rscore>=score[k][l]) {
                    nbetter++;
                } 
                ntotal++;
                n++;
                strcpy(a1->s[w1],orig1);
                strcpy(a2->s[w2],orig2);
         
             }
             ave = (float)sum/(float)n;
             sd =  sqrt( (sumsq - ((sum * sum)/(float)n))/((float)(n-1)));
             /* spec_rand_sd=sqrt((sumsq2-(sum2*sum2)/(float)NRAND2) /((float)(NRAND2-1))); */
            }
            Z = (score[k][l] - ave)/sd;
            if((par->rand==2) || (par->rand==3) || (par->rand==4)) {
               pvalue = (float)nbetter/(float)ntotal;
               min_pvalue = 1/(float)par->nrand;
               if(pvalue<min_pvalue) { pvalue = min_pvalue; }
            } else {
               pvalue = 1.0;
            }
            if(par->diff==1) {
               ref_Z = (0 - ave)/sd;
            } else {
               ref_Z = (ref_score - ave)/sd;
            }
            printf(" rand %8.3f %8.3f %8.3f %7.5f %7.5f %7.5f",ave,sd,Z,pvalue,ref_score,ref_Z); 
           }
           printf("\n");
           fflush(stdout);
           if((np==1) && (par->disp==1)) {
              printf("Exiting after first pair for display mode\n");
              exit(-1);
           }
        }
      }
     }
  }
  exit(0);
}
