#!/usr/bin/perl -w

use strict;

use POSIX 'setsid';

use CGI;
use IO::String;
use Bio::SeqIO;
use Bio::AlignIO;
use Bio::Seq;
use Bio::SearchIO;
use Bio::Graphics;
use Mail::Sendmail;

use TransDB::PDBFile;
use TransDB::Utils;
use Web;

#############################
# Site-specific adjustments #
#############################

# Change the following environment variables if not defined:

# the directory containing your blast matrices
$ENV{'BLASTMAT'} or ($ENV{'BLASTMAT'} = '/g/russell1/apps/blast/data/');

# the directory containing your PDB files
$ENV{'RASMOLPDBPATH'} or ($ENV{'RASMOLPDBPATH'} = '/g/data/pdb/');

# the directory containing your STAMP definition files
$ENV{'STAMPDIR'} or ($ENV{'STAMPDIR'} = '/g/russell1/apps/stamp.4.3/defs/');

# the directory containing your InterPreTS data files
$ENV{'I2DIR'} or ($ENV{'I2DIR'} = '/g/russell1/apps/interprets2/');

# Give the location of your blast pdbaa database
my $blastdb = $ENV{'DS'} . '/blastdb/pdbaa';

# Give the full path of the directory where you want your
# results to be written. If running from the command line,
# this path can be overridden by specifying 'output=path'
# on the command line where path is the desired output path
my $tmp_dn = '/tmp/';

# The following are not important if you are only
# using this script from the command line.

# Give the relative URL to the results directory
my $tmp_dn_url = '/tools/tmp/';

# Who should notification emails come from?
my $email = 'interprets@russelllab.org';

#####################################################
# You should not need to change anything below here #
#####################################################

$SIG{CHLD} = 'IGNORE'; # to avoid zombies in the system process table

my $cgi = CGI->new();

(!$ENV{'GATEWAY_INTERFACE'}) and defined($cgi->param('output')) and ($tmp_dn = $cgi->param('output'));

# Note - head and tail done in multiple places, since head shouldn't be output when redirecting

if(defined($cgi->param('seq_text')) or defined($cgi->param('seq_fn'))) {
    process($cgi);
}
else {
    prompt($cgi);
}

sub prompt {
    my($cgi) = @_;

    Web::head($cgi, 'InterPreTS - Interaction Prediction through Tertiary Structure', 'interprets');

    print <<END;
<div id="content">
<div class="post">
<div class="bgtop">
<div class="bgbtm">

<div id="title">

<h2>InterPreTS - Interaction Prediction through Tertiary Structure</h2>

<p>
Given a set of protein sequences (in fasta format), this tool will use <a href="http://en.wikipedia.org/wiki/BLAST">BLAST</a> to find
homologues of known structure for all pairs (i.e. templates that can model each pair of sequences based on homology)
and then evaluate the suitability of those
templates for modelling the interaction using the
method described in
<a href="http://www.ncbi.nlm.nih.gov/sites/entrez?cmd=Retrieve&amp;db=PubMed&amp;list_uids=11972061&amp;dopt=Abstract">Aloy &amp; Russell, 2002</a>
and
<a href="http://www.ncbi.nlm.nih.gov/sites/entrez?cmd=Retrieve&amp;amp;db=PubMed&amp;list_uids=12499311&amp;dopt=Abstract">Aloy &amp; Russell, 2003</a>.
<!-- More information is available <a href="/interprets2/info.html">here</a>. -->
</p>

</div>

<h3>Download</h3>
<p>
End-Users at non-profit public funded Academic, Education or Research Institutions, please read
the <a href="/tools/interprets_license.pdf">ACADEMIC Software License Agreement</a>. If you
agree, click <a href="/tools/interprets.tar.gz">I AGREE</a> to download.
If you are not a member of a public funded Academic and/or Education
and/or Research Institution you must obtain a commercial license from
<a href="http://www.embl-em.de/">EMBLEM</a>.
</p>

</div> <!-- end of bgbtm div -->
</div> <!-- end of bgtop div -->
</div> <!-- end of post div -->
</div> <!-- end of content div -->
END

    form($cgi);

    Web::tail($cgi, 'interprets'),
}

sub form {
    my($cgi) = @_;

    print(
	  "<div class=\"sidebar\">\n",
	  "<div class=\"bgtop\">\n",
	  "<div class=\"bgbtm\">\n",

	  $cgi->start_multipart_form(),

	  "<ul>\n",
	  "<li>Sequences in <a href=\"http://en.wikipedia.org/wiki/FASTA_format\">FASTA format</a>:",
          '<li>', $cgi->textarea('-id' => 'seq_text', '-name' => 'seq_text', '-default' => '', '-rows' => 5, '-columns' => '20'), "</li>\n",
	  "<li>Or upload a file:</li>\n",
	  '<li>', $cgi->filefield('-id' => 'seq_fn', '-name' => 'seq_fn', '-default' => '', '-size' => 10), "</li>\n",
	  "</ul>\n",
	  "<ul>\n",
	  '<li>BLAST e-value &le;: ', $cgi->textfield('-id' => 'e_value', '-name' => 'e_value', '-default' => 0.01, '-size' => 5), "</li>\n",
	  "<li>Email (optional):</li>\n",
	  '<li>', $cgi->textfield('-id' => 'email', '-name' => 'email'), "</li>\n",
	  "<li>Job name (optional):</li>\n",
	  '<li>', $cgi->textfield('-id' => 'name', '-name' => 'name'), "</li>\n",
	  '<li>', $cgi->submit('-name' => 'Search'), ' ', $cgi->reset(), "</li>\n",
          "</ul>\n",

	  $cgi->end_form, "\n",

	  "</div>\n",
	  "</div>\n",
	  "</div>\n",
	 );
}

sub process {
    my($cgi) = @_;

    my $name;
    my $e_value;
    my $query;
    my $pairs;
    my $pair;
    my $pid;
    my $output;
    my $output_fn;
    my $output_url;
    my $output_fh;
    my $error;
    my $ret;
    my $mail;

#    my $params = $cgi->Vars;
#    my $param;
#    my $value;
#    while(($param, $value) = each %{$params}) {
#	print "'$param' => '$value'<br/>\n";
#    }

    $name = $cgi->param('name') ? $cgi->param('name') : '';

    $e_value = $cgi->param('e_value');
    (defined($e_value) and ($e_value =~ /^([+-]?)(?=\d|\.\d)\d*(\.\d*)?([Ee]([+-]?\d+))?$/)) or ($e_value = 0.01); # ensure e_value is numeric

    $query = {
	      'job_id'   => Web::job_id($cgi, $tmp_dn),
	      'e_value'  => $e_value,
	      'seq_fn'   => undef,
	      'seqs'     => {},
	      'blast_fn' => undef,
	      'pairs'    => {},
	      'n_pairs'  => 0,
	     };

    if(!mkdir(join('', $tmp_dn, $query->{'job_id'}))) {
	print(
	      $cgi->header(), "\n", $cgi->start_html(), "\n",
	      Web::error_string($cgi, (join('', "Could not create '", $tmp_dn, $query->{'job_id'}, "' directory"))),
	      $cgi->end_html(), "\n",
	     );
    }
    else {
	$tmp_dn = join('', $tmp_dn, $query->{'job_id'}, '/');
	$tmp_dn_url = join('', $tmp_dn_url, $query->{'job_id'}, '/');

	$output_fn = $tmp_dn . 'index.html';
	$output_url = join('', 'http://', $ENV{'SERVER_NAME'}, $tmp_dn_url, 'index.html');

	# fork and wait for the output
	select(STDOUT);
	$| = 1;
	$pid = fork();

	if(!defined($pid)) { # can't fork
	    print(
		  $cgi->header(), "\n", $cgi->start_html(), "\n",
		  Web::error_string($cgi, 'Could not fork'),
		  $cgi->end_html(), "\n",
		 );
	}
	elsif($pid == 0) { # child process
	    # close the following so that the parent isn't
	    # kept alive waiting for the child to finish
	    close(STDIN);
	    close(STDOUT);
	    close(STDERR);

	    if(!setsid) {
		Web::error($cgi, 'Could not set session ID'),
		exit(0);
	    }

	    $error = '';

	    $ret = blast($cgi, $query);
	    ($ret->[0] != 0) and ($error .= $ret->[1]);

	    # ret == 0 - ok
	    # ret == 1 - fatal error
	    # ret >= 2 - non-fatal error

	    if($ret->[0] != 1) {
		$ret = process_blast($cgi, $query);
		($ret->[0] != 0) and ($error .= $ret->[1]);
	    }

	    if($ret->[0] != 1) {
		$ret = pairs2doms($cgi, $query);
		($ret->[0] != 0) and ($error .= $ret->[1]);
	    }

	    if($ret->[0] != 1) {
		$ret = write_doms($cgi, $query);
		($ret->[0] != 0) and ($error .= $ret->[1]);
	    }

	    if($ret->[0] != 1) {
		$ret = n_contacts($cgi, $query);
		($ret->[0] != 0) and ($error .= $ret->[1]);
	    }

	    if($ret->[0] != 1) {
		$ret = doms2seqs($cgi, $query);
		($ret->[0] != 0) and ($error .= $ret->[1]);
	    }

	    if($ret->[0] != 1) {
		$ret = align2templates($cgi, $query);
		($ret->[0] != 0) and ($error .= $ret->[1]);
	    }

	    if($ret->[0] != 1) {
		$ret = run_interprets($cgi, $query);
		($ret->[0] != 0) and ($error .= $ret->[1]);
	    }

	    if($ret->[0] != 1) {
		$ret = write_jmol($cgi, $query);
		($ret->[0] != 0) and ($error .= $ret->[1]);
	    }

	    $output = output_pairs($cgi, $query);

	    # print this to a file
	    open($output_fh, ">$output_fn") or Web::error($cgi, "Could not open '$output_fn' file for writing.");
	    Web::head($cgi, sprintf("InterPreTS job '%s' finished", $name ? $name : $query->{'job_id'}), 'interprets', $output_fh);
	    print(
		  $output_fh
                  "<div class=\"post\">\n",
                  "<div class=\"bgtop\">\n",
                  "<div class=\"bgbtm\">\n",
		  "<h2>InterPreTS</h2>\n",
		  "<p>Your job ", ($name ? "'$name' " : ''), " has finished and the results are below.",
		  " These will be deleted automatically after seven days.</p>\n",
		  ($error ne '') ? Web::error_string($cgi, $error) : '',
		  "<div id=\"results\" class=\"results\">\n",
		  $output,
		  "</div>\n",
                  "</div> <!-- end of bgbtm div -->\n",
                  "</div> <!-- end of bgtop div -->\n",
                  "</div> <!-- end of post div -->\n",
		 );
	    Web::tail($cgi, 'interprets', $output_fh);
	    close($output_fh);

	    # FIXME - what if printing of the above has started but
	    # not finished when the browser is refreshing the page?

	    if($cgi->param('email')) {
		$mail = {
			 'To'      => $cgi->param('email'),
			 'From'    => $email,
			 'Message' => $name ? "Your InterPreTS results for '$name' are ready at $output_url" : "Your InterPreTS results are ready at $output_url",
			 'Subject' => $name ? "InterPreTS Results for '$name'" : 'InterPreTS Results',
			};

		sendmail(%{$mail}); # FIXME - error trap this
	    }


	    # exit so that the child doesn't stray into code for the parent
	    exit(0);
	}
	else { # parent process
	    # create the page that tells the users to wait (over-written with results later)
	    open($output_fh, ">$output_fn") or Web::error($cgi, "Could not open '$output_fn' file for writing.");
	    Web::head($cgi,
		      sprintf("InterPreTS job '%s' in progress", $name ? $name : $query->{'job_id'}),
		      'interprets',
		      $output_fh,
		      $cgi->meta({'-http_equiv' => 'REFRESH', '-content' => 10}),
		     );
	    print(
		  $output_fh
                  "<div class=\"post\">\n",
                  "<div class=\"bgtop\">\n",
                  "<div class=\"bgbtm\">\n",
		  "<h2>InterPreTS</h2>\n",
		  "<p>\n",
		  "InterPreTS is running. ",
		 );

	    if($cgi->param('email')) {
		print(
		      $output_fh
		      "An email will be sent to ", $cgi->param('email') , " when your job is finished. ",
		      "This page will refresh every ten seconds until then. ",
		      "Please either leave this page open, bookmark it and check again later, or wait for the email.",
		     );
	    }
	    else {
		print(
		      $output_fh
		      "This page will refresh every ten seconds until your job has finished. ",
		      "Please either leave this page open or bookmark it and check again later.",
		     );
	    }

	    print(
		  $output_fh
		  "</p>\n",
                  "</div> <!-- end of bgbtm div -->\n",
                  "</div> <!-- end of bgtop div -->\n",
                  "</div> <!-- end of post div -->\n",
                 );

	    Web::tail($cgi, 'interprets', $output_fh);
	    close($output_fh);

	    # redirect the browser to the wait page
	    print $cgi->redirect('-url' => $output_url), "\n";

	    exit(0);
	}
    }
}

sub blast {
    my($cgi, $query) = @_;

    my $cgi_error;
    my $fh;
    my $e_value;
    my $seq_text;
    my $seq_fn;
    my $seq_tmp_fn;
    my $seq_fh;
    my $seq_line;
    my $blast_prog;
    my $blast_fn;
    my $cmd;
    my $pairs;
    my $in;
    my $seq;
    my $error;

    #print "# blast()<br/>\n";

    # FIXME - implement a queue

    # check that blast is available
    if(!open($fh, "blastall |")) {
	$error = [1, "Cannot run 'blastall' program.<br/>\n"];
	return $error;
    }
    close($fh);

    # process the input
    $e_value = $query->{'e_value'};

    $seq_text = $cgi->param('seq_text');
    if(defined($cgi_error = $cgi->cgi_error)) {
	$error = [1, $cgi_error];
	return $error;
    }

    $seq_fn = $cgi->param('seq_fn');
    if(defined($cgi_error = $cgi->cgi_error)) {
	$error = [1, $cgi_error];
	return $error;
    }

    # run blast on each sequence

    # get the query sequences
    if(defined($seq_text) and ($seq_text ne '')) {
	$fh = IO::String->new($seq_text);
    }
    elsif(defined($seq_fn) and ($seq_fn ne '')) {
	# FIXME - could be better to check 'CGI or not' elsewhere,
	# but the only difference for now is with file uploads.

	if($ENV{'GATEWAY_INTERFACE'}) { # is being run by apache or somesuch
	    if(!defined($fh = $cgi->upload('seq_fn'))) {
		$error = [1, join('', "'", $cgi->param('seq_fn') , "' is not a valid filehandle")];
		return $error;
	    }
	}
	else {# is being run from the command line
	    if(!open($fh, $cgi->param('seq_fn'))) {
		$error = [1, join('', "cannot open '", $cgi->param('seq_fn') , "' for reading")];
		return $error;
	    }
	}
    }
    else {
	# FIXME - verify that a sequence has been entered before allowing 'submit'
	$error = [1, "No sequences read.<br/>\n"];
	return $error;
    }

    # write the query sequences to a file so that blast can use them
    $seq_tmp_fn = $tmp_dn . 'query.aa';
    $query->{'seq_fn'} = $seq_tmp_fn;
    if(!open($seq_fh, ">$seq_tmp_fn")) {
	$error = [1, "Cannot open '$seq_tmp_fn' file for writing.<br/>\n"];
	return $error;
    }
    while($seq_line = <$fh>) {
	print $seq_fh $seq_line;
    }
    close($seq_fh);

    # read in Bio::Seq objects
    $in = Bio::SeqIO->new('-file' => $seq_tmp_fn, '-format' => 'fasta');
    while($seq = $in->next_seq) {
	$query->{'seqs'}->{$seq->id} = $seq;
    }

    # FIXME - check the number of sequences
    # FIXME - have a max number of sequences allowed

    # run blast, writing results to a temporary file
    $blast_prog = (defined($cgi->param('seq_type')) and $cgi->param('seq_type') eq 'dna') ? 'blastx' : 'blastp';
    $blast_fn  = $tmp_dn . $blast_prog . '.txt';
    $query->{'blast_fn'} = $blast_fn;
    $cmd = "blastall -p $blast_prog -d $blastdb -e $e_value -i $seq_tmp_fn";
    qx($cmd 1> $blast_fn 2> ${blast_fn}.err);  # FIXME - error trap this

    return [0, ''];
}

sub process_blast {
    my($cgi, $query) = @_;

    my $fh;
    my @errors;
    my @query_ids;
    my $i;
    my $j;
    my $in;
    my $hits;
    my $desc;
    my $name;
    my $result;
    my $hit0;
    my $hit1;
    my $hsp0;
    my $hsp1;
    my $pdb_id;
    my $chain_id0;
    my $chain_id1;
    my $overlap;
    my $n_pairs;
    my $n_pairs_old;

    #print "# process_blast()<br/>\n";

    if(!open($fh, join('', $query->{'blast_fn'}, '.err'))) {
	return [1, join('', "cannot open '", $query->{'blast_fn'}, ".err' file for reading")];
    }
    @errors = <$fh>;
    close($fh);
    (@errors > 0) and return([1, join('', @errors)]);

    # store the results by PDB code
    $in = Bio::SearchIO->new('-file' => $query->{'blast_fn'}, '-format' => 'blast');
    $hits = {};
    while($result = $in->next_result) {
	#printf "## query = '%s'<br/>\n", $result->query_name;

	$hits->{$result->query_name} = {};
	while($hit0 = $result->next_hit) {
	    ($pdb_id, $chain_id0) = parse_hit_name($cgi, $hit0->name);

	    defined($hits->{$pdb_id}) or ($hits->{$pdb_id} = {});
	    defined($hits->{$pdb_id}->{$result->query_name}) or ($hits->{$pdb_id}->{$result->query_name} = {});
	    $hits->{$pdb_id}->{$result->query_name}->{$chain_id0} = $hit0;

	    # parse out the redundancies too
	    $desc = $hit0->description;
	    while($desc =~ /(pdb\|\S+\|\S*)/g) {
		$name = $1;
		($pdb_id, $chain_id0) = parse_hit_name($cgi, $name);

		defined($hits->{$pdb_id}) or ($hits->{$pdb_id} = {});
		defined($hits->{$pdb_id}->{$result->query_name}) or ($hits->{$pdb_id}->{$result->query_name} = {});
		$hits->{$pdb_id}->{$result->query_name}->{$chain_id0} = $hit0;
	    }
	}
    }

    # look at all pairs of queries where both queries hit the same PDB
    $n_pairs = 0;
    $n_pairs_old = 0;
    $query->{'pairs'} = {};
    foreach $pdb_id (keys %{$hits}) {
	@query_ids = keys %{$hits->{$pdb_id}};
	(@query_ids < 2) and next;
	for($i = 0; $i < @query_ids; $i++) {
	    for($j = $i + 1; $j < @query_ids; $j++) {
		foreach $chain_id0 (keys %{$hits->{$pdb_id}->{$query_ids[$i]}}) {
		    $hit0 = $hits->{$pdb_id}->{$query_ids[$i]}->{$chain_id0};
		    foreach $chain_id1 (keys %{$hits->{$pdb_id}->{$query_ids[$j]}}) {
			$hit1 = $hits->{$pdb_id}->{$query_ids[$j]}->{$chain_id1};

			if($chain_id1 eq $chain_id0) { # same chain hit by both queries
			    # FIXME - if the hits are on the same chain, do they also overlap?
			    foreach $hsp0 ($hit0->hsps) {
				foreach $hsp1 ($hit1->hsps) {
				    $overlap = overlap($hsp0->start('subject'), $hsp0->end('subject'), $hsp1->start('subject'), $hsp1->end('subject'));
				    if($overlap > 0) {
					#print "BAD: $pdb_id|$chain_id0 - same pdb, same chain, overlap of $overlap<br/>\n";
				    }
				    else {
					#print "GOOD: $pdb_id|$chain_id0 - same pdb, same chain, no overlap<br/>\n";

					# these hsps are checked later for contacts, after the dom files have been
					# generated, as this requires the pdbaa sequence coordinates to be translated
					# in to PDB res ids

					defined($query->{'pairs'}->{$query_ids[$i]}) or ($query->{'pairs'}->{$query_ids[$i]} = {});
					defined($query->{'pairs'}->{$query_ids[$i]}->{$query_ids[$j]})
					  or ($query->{'pairs'}->{$query_ids[$i]}->{$query_ids[$j]} = []);

					push @{$query->{'pairs'}->{$query_ids[$i]}->{$query_ids[$j]}},
					  {
					   'error'     => 0,
					   'id'        => $n_pairs++,
					   'pdb'       => {'id' => $pdb_id, 'classification' => '', 'title' => ''},
					   'dom_ids'   => [$pdb_id . $chain_id0 . 'a', $pdb_id . $chain_id1 . 'b'],
					   'chain_ids' => [$chain_id0, $chain_id1],
					   'hits'      => [$hit0, $hit1],
					   'hsps'      => [$hsp0, $hsp1],
					   'res'       => [], # to be filled in with pdb residue IDs later
					   'ins'       => [], # to be filled in with pdb residue inserstion codes later
					   'doms'      => [], # to be filled in with stamp format domains later
					   'dom_fn'    => undef,
					   'contacts'  => 0, # filled in later by running interprets -count on the dom file
					   'domseqs'   => [],
					   'aln_fns'   => [],
					   'alns'      => [],
					   'i2'        => {
							   'fn'      => '',
							   'ras_fn'  => '',
							   'jmol_fn' => '',
							   'raw'     => -100,
							   'mean'    => -100,
							   'sd'      => -100,
							   'z_score' => -100,
							  },
					  };
					$query->{'n_pairs'}++;
				    }
				}
			    }
			}
			else {
			    #print "GOOD: $pdb_id|$chain_id0|$chain_id1 - same pdb, different chains<br/>\n";
			    foreach $hsp0 ($hit0->hsps) {
				foreach $hsp1 ($hit1->hsps) {
				    # these hsps are checked later for contacts, after the dom files have been
				    # generated, as this requires the pdbaa sequence coordinates to be translated
				    # in to PDB res ids

				    defined($query->{'pairs'}->{$query_ids[$i]}) or ($query->{'pairs'}->{$query_ids[$i]} = {});
				    defined($query->{'pairs'}->{$query_ids[$i]}->{$query_ids[$j]})
				      or ($query->{'pairs'}->{$query_ids[$i]}->{$query_ids[$j]} = []);

				    push @{$query->{'pairs'}->{$query_ids[$i]}->{$query_ids[$j]}},
				      {
				       'error'     => 0,
				       'id'        => $n_pairs++,
				       'pdb'       => {'id' => $pdb_id, 'classification' => '', 'title' => ''},
				       'dom_ids'   => [$pdb_id . $chain_id0 . 'a', $pdb_id . $chain_id1 . 'b'],
				       'chain_ids' => [$chain_id0, $chain_id1],
				       'hits'      => [$hit0, $hit1],
				       'hsps'      => [$hsp0, $hsp1],
				       'res'       => [], # to be filled in with pdb residue IDs later
				       'ins'       => [], # to be filled in with pdb residue inserstion codes later
				       'doms'      => [], # to be filled in with stamp format domains later
				       'dom_fn'    => undef,
				       'contacts'  => 0, # filled in later by running interprets -count on the dom file
				       'domseqs'   => [],
				       'aln_fns'   => [],
				       'alns'      => [],
				       'i2'        => {
						       'fn'      => '',
						       'ras_fn'  => '',
						       'jmol_fn' => '',
						       'raw'     => -100,
						       'mean'    => -100,
						       'sd'      => -100,
						       'z_score' => -100,
						      },
				      };
				    $query->{'n_pairs'}++;
				}
			    }
			}
		    }
		}
		($n_pairs != $n_pairs_old) and
		  ($query->{'pairs'}->{$query_ids[$i]}->{$query_ids[$j]} = [sort _by_evalue @{$query->{'pairs'}->{$query_ids[$i]}->{$query_ids[$j]}}]);
		$n_pairs_old = $n_pairs;
	    }
	}
    }

    return [0, ''];
}

sub _by_evalue {
    my $e_a;
    my $e_b;
    my $cmp;

    $e_a = ($a->{'hsps'}->[0]->evalue < $a->{'hsps'}->[1]->evalue) ? $a->{'hsps'}->[0]->evalue : $a->{'hsps'}->[1]->evalue;
    $e_b = ($b->{'hsps'}->[0]->evalue < $b->{'hsps'}->[1]->evalue) ? $b->{'hsps'}->[0]->evalue : $b->{'hsps'}->[1]->evalue;

    $cmp = $e_a <=> $e_b;

    return $cmp;
}

sub _by_z {
    my $cmp;

    if(defined($a->{'i2'})) {
	if(defined($b->{'i2'})) {
	    $cmp = $b->{'i2'}->{'z_score'} <=> $a->{'i2'}->{'z_score'};
	}
	else {
	    $cmp = -1;
	}
    }
    elsif(defined($b->{'i2'})) {
	$cmp = 1;
    }
    else {
	$cmp = $a->{'hsps'}->[0]->evalue <=> $b->{'hsps'}->[0]->evalue;
	($cmp == 0) and ($cmp = $a->{'hsps'}->[1]->evalue <=> $b->{'hsps'}->[1]->evalue);
    }

    return $cmp;
}

sub overlap {
    my($start0, $end0, $start1, $end1) = @_;

    # FIXME - I think there's a bioperl function somewhere for calculating overlaps...

    my $overlap;

    if($start0 <= $start1) {
	if($end0 >= $start1) {
	    if($end0 >= $end1) {
		# ------------
		#     ----
		#
		# $hsp0 contains $hsp1

		$overlap = $end1 - $start1 + 1;
	    }
	    else {
		# --------
		#     --------
		#
		# the end of $hsp0 overlaps with the beginning of $hsp1

		$overlap = $end0 - $start1 + 1;
	    }
	}
	else {
	    # ----
	    #       ----
	    #
	    # $hsp0 is before $hsp1
	    $overlap = 0;
	}
    }
    else {
	if($start0 <= $end1) {
	    if($end0 <= $end1) {
		#     ----
		# ------------
		#
		# $hsp0 is contained by $hsp1

		$overlap = $end0 - $start0 + 1;
	    }
	    else {
		#     --------
		# --------
		#
		# the beginning of $hsp0 overlaps with the end of $hsp1

		$overlap = $end1 - $start0 + 1;
	    }
	}
	else {
	    # $hsp0 is after $hsp1
	    $overlap = 0;
	}
    }

    return $overlap;
}

sub parse_hit_name {
    my($cgi, $hit_name) = @_;

    my $pdb_id;
    my $chain_id;

    if($hit_name =~ /pdb\|(\S+)\|(.*)/) {
	($pdb_id, $chain_id) = ($1, $2);
	$pdb_id =~ tr/[A-Z]/[a-z]/;
	#$chain_id = ($chain_id =~ /\A\S+-(\S+)/) ? $1 : '_';
	if($chain_id eq '') {
	    $chain_id = '_';
	}
	elsif(length($chain_id) == 2) {
	    # lowercase chain IDs are represented by double uppercase, eg '1WF4|AA' = '1wf4 chain a', for some reason
	    ($chain_id = substr($chain_id, 0, 1)) =~ tr/[A-Z]/[a-z]/;
	}
    }
    else {
	Web::error($cgi, "Cannot parse hit name '$hit_name'");
    }

    return($pdb_id, $chain_id);
}

sub parse_pdb {
    my($cgi, $pdb_id, $fh) = @_;

    my $pdb;
    my $chain;
    my $previous_num;
    my @data;
    my $alt;
    my $code3;
    my $chain_id;
    my $num;
    my $ins;
    my $previous_ins;
    my $previous_chain_id;
    my $code1;

    my %three2one = (
		     'ALA' => 'A',
		     'CYS' => 'C',
		     'ASP' => 'D',
		     'GLU' => 'E',
		     'PHE' => 'F',
		     'GLY' => 'G',
		     'HIS' => 'H',
		     'ILE' => 'I',
		     'LYS' => 'K',
		     'LEU' => 'L',
		     'MET' => 'M',
		     'ASN' => 'N',
		     'PRO' => 'P',
		     'GLN' => 'Q',
		     'ARG' => 'R',
		     'SER' => 'S',
		     'THR' => 'T',
		     'VAL' => 'V',
		     'TRP' => 'W',
		     'TYR' => 'Y',
		    );

    $pdb = {'classification' => '', 'title' => '', 'chains' => {}};
    $chain = {'id' => '_', 'aas' => '', 'res_ids' => ['beg']}; # 'beg' because sequence coordinates are 1-based
    $previous_num = -1000000;
    $previous_ins = '';
    $previous_chain_id = '';
    while(<$fh>) {
	/^ENDMDL/ and last;

	if(/^HEADER.{4}(.{1,40})/) {
	    $pdb->{'classification'} = $1;
	}
	elsif(/^TITLE.{4}(.{1,60})/) {
	    $pdb->{'title'} .= $1;
	}
	elsif(/^ATOM.{9}CA/) {
	    # FIXME - deal with insertion codes

	    @data = unpack 'a6a5a5aa3a2a4a', $_;
	    ($alt      = $data[3]) =~ s/\s+//g;
	    ($code3    = $data[4]) =~ s/\s+//g;
	    ($chain_id = $data[5]) =~ s/\s+//g;
	    ($num      = $data[6]) =~ s/\s+//g;
	    ($ins      = $data[7]) =~ s/\s+//g;

	    ($chain_id eq '') and ($chain_id = '_');
	    ($ins eq '') and ($ins = '_');

	    if($chain_id ne $previous_chain_id) {
		if($previous_chain_id ne '') {
		    $chain->{'id'} = $previous_chain_id;
		    $pdb->{'chains'}->{$chain->{'id'}} = $chain;
		}

		$previous_num = -1000000;
		$previous_ins = '';
		$chain = {'id' => '_', 'aas' => '', 'res_ids' => ['beg'], 'ins' => ['beg']}; # 'beg' because sequence coordinates are 1-based
	    }

	    if(($num != $previous_num) or ($ins ne $previous_ins)) {
		$code1 = defined($three2one{$code3}) ? $three2one{$code3} : 'X';
		$chain->{'aas'} .= $code1;
		push @{$chain->{'res_ids'}}, $num;
		push @{$chain->{'ins'}}, $ins;
	    }

	    $previous_num = $num;
	    $previous_ins = $ins;
	    $previous_chain_id = $chain_id;
	}
    }
    close($fh);

#    warn "$pdb_id\n";
#    foreach $chain (values %{$pdb->{'chains'}}) {
#	for(my $i = 0; $i < @{$chain->{'res_ids'}}; $i++) {
#	    warn($chain->{'res_ids'}->[$i], ' ', $chain->{'ins'}->[$i], "\n");
#	}
#    }

    $pdb->{'title'} =~ s/\A\s*(.*?)\s*\Z/$1/;
    $pdb->{'classification'} =~ s/\A\s*(.*?)\s*\Z/$1/;

    $chain->{'id'} = $chain_id;
    $pdb->{'chains'}->{$chain->{'id'}} = $chain;

    foreach $chain (values %{$pdb->{'chains'}}) {
	$chain->{'seq'} = Bio::Seq->new('-id' => join($pdb_id, $chain->{'id'}), '-seq' => $chain->{'aas'});
    }

    return $pdb;
}

sub pairs2doms {
    my($cgi, $query) = @_;

    my $query_id0;
    my $query_id1;
    my $pair;
    my $pdb_fn;
    my $pdb;
    my $pdb_id;
    my $chain_id;
    my $chain;
    my $atom_seq;
    my $seq_in;
    my $seq_out;
    my $seq_fn;
    my $fasta_seq;
    my $cmd;
    my $fh;
    my $aln_in;
    my $aln;
    my $seq0;
    my $seq1;
    my $i;
    my $j;
    my $pos0;
    my $pos1;
    my $res_id;
    my $aln_fn;
    my $pos2res;
    my $hsp;
    my $res0;
    my $res1;
    my $ins0;
    my $ins1;
    my $db_id;
    my $error;

    # FIXME - use pre-computed data, i.e. pdbaa to pdb ATOM seq alignments, etc

    #print "# pairs2doms()<br/>\n";

    $error = [0, ''];
    foreach $query_id0 (keys %{$query->{'pairs'}}) { 
	foreach $query_id1 (keys %{$query->{'pairs'}->{$query_id0}}) { 
	    PAIR: foreach $pair (@{$query->{'pairs'}->{$query_id0}->{$query_id1}}) { 
		($pair->{'error'} != 0) and next;

		$pdb_id = $pair->{'pdb'}->{'id'};

		# parse the pdb

		$pdb_fn = TransDB::PDBFile::get_fn($pdb_id);

		if(!defined($fh = TransDB::Utils::openfile($pdb_fn))) {
		    $pair->{'error'}++;
		    $error->[0] = 2;
		    $error->[1] .= "Cannot open '$pdb_fn' file for reading.<br/>\n";
		    next PAIR;
		}
		$pdb = parse_pdb($cgi, $pdb_id, $fh);
		close($fh);

		# save some of the pdb info (but not the atoms etc., to save memory)
		$pair->{'pdb'}->{'classification'} = $pdb->{'classification'};
		$pair->{'pdb'}->{'title'} = $pdb->{'title'};

		for($i = 0; $i < 2; $i++) {
		    $hsp = $pair->{'hsps'}->[$i];
		    $chain_id = $pair->{'chain_ids'}->[$i];
		    $chain = $pdb->{'chains'}->{$chain_id};

		    # get the sequence parsed from the pdb
		    if(defined($pdb->{'chains'}->{$chain_id})) {
			$atom_seq = $pdb->{'chains'}->{$chain_id}->{'seq'};
		    }
		    else {
			$pair->{'error'}++;
			$error->[0] = 2;
			$error->[1] .= "Chain ID '$chain_id' not found in pdb '$pdb_id'<br/>\n";
			next PAIR;
		    }

		    # get the sequence from the blast db
		    if($chain_id eq '_') {
			$db_id = '';
		    }
		    elsif($chain_id =~ /[a-z]/) {
			# lowercase chain IDs are represented by double uppercase, eg '1WF4|AA' = '1wf4 chain a', for some reason
			$db_id = $chain_id . $chain_id;
		    }
		    else {
			$db_id = $chain_id;
		    }
		    $db_id = join('|', $pdb_id, $db_id);
		    $db_id =~ tr/[a-z]/[A-Z]/;
		    $db_id = 'pdb|' . $db_id;
		    $seq_fn = join('', $tmp_dn, $i, '.', $pair->{'id'}, '.pdbseq.aa');
		    $cmd = "fastacmd -d $blastdb -s '$db_id' > $seq_fn";

		    if(Web::mysystem($cmd) != 0) {
			$pair->{'error'}++;
			$error->[0] = 1;
			$error->[1] .= "'$cmd' failed: '$!'<br/>\n";
			next PAIR;
		    }

		    $seq_in = Bio::SeqIO->new('-file' => $seq_fn, '-format' => 'fasta');
		    $fasta_seq = $seq_in->next_seq();

		    # align the two
		    $seq_fn = join('', $tmp_dn, $i, '.', $pair->{'id'}, '.pdb.aa');
		    $seq_out = Bio::SeqIO->new('-file' => ">$seq_fn", '-format' => 'fasta');
		    $seq_out->write_seq($fasta_seq, $atom_seq);
		    $aln_fn = join('', $tmp_dn, $i, '.', $pair->{'id'}, '.pdb.muscle');

		    #$cmd = "muscle -quiet -stable -clwstrict -in $seq_fn -out $aln_fn";
		    $cmd = "muscle -quiet -clwstrict -in $seq_fn -out $aln_fn"; # -stable not supported in muscle 3.8


		    if(Web::mysystem($cmd) != 0) {
			$pair->{'error'}++;
			$error->[0] = 1;
			$error->[1] .= "'$cmd' failed: '$!'<br/>\n";
			next PAIR;
		    }

		    $aln_in = Bio::AlignIO->new('-file' => $aln_fn, '-format' => 'clustalw');
		    $aln = $aln_in->next_aln();

		    # convert from blast db sequence coordinates to pdb residues IDs
		    ($seq0, $seq1) = $aln->each_seq;

		    # can not depend on sequences in alignment being in a particular order without muscle -stable
		    ($seq0->id eq $fasta_seq->id) or (($seq0, $seq1) = ($seq1, $seq0));

		    $pos0 = 0;
		    $pos1 = 0;
		    $pos2res = {};
		    #print "<pre>\n";
		    for($j = 1; $j <= $aln->length; $j++) {
			# convert from blast db sequence coordinates to pdb sequence coordinates
			($seq0->subseq($j, $j) eq '-') or $pos0++;
			($seq1->subseq($j, $j) eq '-') or $pos1++;

			# convert from pdb sequence coordinates to pdb residue IDs
			$res_id = $chain->{'res_ids'}->[($pos1 == 0) ? 1 : $pos1];
			$ins0 = $chain->{'ins'}->[($pos1 == 0) ? 1 : $pos1];

			# store the conversion from blast db sequence coordinates to pdb residue IDs
			$pos2res->{$pos0} = [$res_id, $ins0];

			#printf "%s%s %03d %03d %03d\n", $seq0->subseq($j, $j), $seq1->subseq($j, $j), ($pos0 == 0) ? 1 : $pos0, ($pos1 == 0) ? 1 : $pos1, $res_id;
		    }
		    #print "\n</pre>\n";

		    # convert the hsp start and end positions in to pdb residue ids
		    ($res0, $ins0, $res1, $ins1) = (@{$pos2res->{$hsp->start('subject')}}, @{$pos2res->{$hsp->end('subject')}});

		    push @{$pair->{'res'}}, [$res0, $res1];
		    push @{$pair->{'ins'}}, [$ins0, $ins1];

		    # stamp format domain info
		    push @{$pair->{'doms'}}, sprintf(
						     "%s %s { %s %d %s to %s %d %s }",
						     $pdb_fn,
						     $pair->{'dom_ids'}->[$i],
						     $chain_id,
						     $res0,
						     $ins0,
						     $chain_id,
						     $res1,
						     $ins1,
						    );
		}
	    }
	}
    }

    return $error;
}

sub output_pairs {
    my($cgi, $query) = @_;

    my $query_id0;
    my $query_id1;
    my $pair;
    my $aln_url0;
    my $aln_url1;
    my $blast_url;
    my $i2_url;
    my $output;
    my @names;
    my $name;
    my $z_score;
    my $jmol_url;
    my $n_in_contact;

    ($blast_url = $query->{'blast_fn'}) =~ s/.*\///;
    $blast_url = $tmp_dn_url . $blast_url;

    $n_in_contact = 0;
    @names = ();
    if($query->{'n_pairs'} > 0) {

	# first sort results for each pair of queries by z_score
	foreach $query_id0 (keys %{$query->{'pairs'}}) {
	    foreach $query_id1 (keys %{$query->{'pairs'}->{$query_id0}}) {
		($query->{'pairs'}->{$query_id0}->{$query_id1} = [sort _by_z @{$query->{'pairs'}->{$query_id0}->{$query_id1}}]);
		push @names, sprintf("%s with %s", $query_id0, $query_id1);
	    }
	}

	$output .= "<p>(<a href=\"$blast_url\">BLAST results</a>.)</p>\n";

	if(@names > 1) {
	    $output .= "<p>\n<h4>Pairs:</h4>\n<ul>\n";
	    foreach $name (@names) {
		$output .= sprintf("<li><a href=\"#%s\">%s</a></li>\n", $name, $name);
	    }
	    $output .= "</ul>\n</p>\n";
	}

	$output .= z_key();

	foreach $query_id0 (keys %{$query->{'pairs'}}) {
	    foreach $query_id1 (keys %{$query->{'pairs'}->{$query_id0}}) {
		$output .= sprintf(
				   "<h4><a name=\"%s with %s\">%s with %s</a></h4>\n",
				   $query_id0,
				   $query_id1,
				   $query_id0,
				   $query_id1,
				  );

		foreach $pair (@{$query->{'pairs'}->{$query_id0}->{$query_id1}}) {
		    ($pair->{'error'} != 0) and next;
		    ($pair->{'contacts'} < 10) and next;

		    $n_in_contact++;

		    ($aln_url0 = $pair->{'aln_fns'}->[0]) =~ s/.*\///;
		    $aln_url0 = $tmp_dn_url . $aln_url0;

		    ($aln_url1 = $pair->{'aln_fns'}->[1]) =~ s/.*\///;
		    $aln_url1 = $tmp_dn_url . $aln_url1;

		    ($i2_url = $pair->{'i2'}->{'fn'}) =~ s/.*\///;
		    $i2_url = $tmp_dn_url . $i2_url;

		    ($jmol_url = $pair->{'i2'}->{'jmol_fn'}) =~ s/.*\///;
		    $jmol_url = $tmp_dn_url . $jmol_url;

		    $output .= sprintf("<p class=\"%s\">\n", z2class($pair->{'i2'}->{'z_score'}));

		    $output .= sprintf(
				       "Z-score: <a href=\"%s\">%.3f</a> (raw = %.3f, &mu; = %.3f, &sigma; = %.3f) <a href=\"%s\">Jmol</a><br/>\n",
				       $i2_url,
				       $pair->{'i2'}->{'z_score'},
				       $pair->{'i2'}->{'raw'},
				       $pair->{'i2'}->{'mean'},
				       $pair->{'i2'}->{'sd'},
				       $jmol_url,
				      );

		    $output .= sprintf(
				       "Template: <a href=\"http://www.rcsb.org/pdb/explore/explore.do?structureId=%s\">%s</a> %s<br/>\n",
				       $pair->{'pdb'}->{'id'},
				       $pair->{'pdb'}->{'id'},
				       $pair->{'pdb'}->{'title'},
				      );

		    $output .= sprintf(
				       "Alignment 1: <a href=\"%s\">Query1/%d-%d on %s/%d-%d</a> (BLAST E = %.1e, %.1f%% id)<br/>\n",
				       $aln_url1,
				       $pair->{'hsps'}->[0]->start('query'),
				       $pair->{'hsps'}->[0]->end('query'),
				       $pair->{'chain_ids'}->[0],
				       @{$pair->{'res'}->[0]},
				       $pair->{'hsps'}->[0]->evalue,
				       $pair->{'hsps'}->[0]->frac_identical('total') * 100,
				      );

		    $output .= sprintf(
				       "Alignment 2: <a href=\"%s\">Query2/%d-%d on %s/%d-%d</a> (BLAST E = %.1e, %.1f%% id)<br/>\n",
				       $aln_url0,
				       $pair->{'hsps'}->[1]->start('query'),
				       $pair->{'hsps'}->[1]->end('query'),
				       $pair->{'chain_ids'}->[1],
				       @{$pair->{'res'}->[1]},
				       $pair->{'hsps'}->[1]->evalue,
				       $pair->{'hsps'}->[1]->frac_identical('total') * 100,
				      );

		    $output .= "</p>\n";
		}
	    }
	}

	if($n_in_contact == 0) {
	    $output = join('',
			   "<h3>Sorry, I didn't find any suitable templates.</h3>\n",
			   "<p>\n",
			   "BLAST found non-overlapping hits in the same pdb but they were not in contact.\n",
			   "The E-value threshold was ", $query->{'e_value'},
			   " (<a href=\"$blast_url\">Raw BLAST results</a>.)\n",
			   "</p>\n",
			  );
	}
    }
    else {
	$output = join('',
		       "<h3>Sorry, I didn't find any suitable templates.</h3>\n",
		       "<p>\n",
		       "BLAST did not find any non-overlapping hits in the same pdb.\n",
		       "The E-value threshold was ", $query->{'e_value'},
		       " (<a href=\"$blast_url\">Raw BLAST results</a>.)\n",
		       "</p>\n",
		      );
    }

    return $output;
}

sub z2class {
    my($z_score) = @_;

    my $class;

    if($z_score >= 3.0) {
	$class = 'great';
    }
    elsif($z_score >= 2.33) {
	$class = 'good';
    }
    elsif($z_score >= 1.65) {
	$class = 'ok';
    }
    elsif($z_score >= 0.0) {
	$class = 'poor';
    }
    else {
	$class = 'bad';
    }

    return $class;
}

sub z_key {
    my $key;
    my $element;
    my $output;

    $key = [
	    ['great', '3.00'],
	    ['good',  '2.33'],
	    ['ok',    '1.65'],
	    ['poor',  '0.00'],
	    ['bad',   '-&infin;'],
	   ];

    $output = "<p>\n<h4>Key</h4>\nZ-score &ge; \n";
    foreach $element (@{$key}) {
	$output .= sprintf("<span class=\"%s\">%s</span>\n", @{$element});
    }
    $output .= "</p>\n";

    return $output;
}

sub write_doms {
    my($cgi, $query) = @_;

    my $query_id0;
    my $query_id1;
    my $pair;
    my $fn;
    my $fh;
    my $error;

    #print "# write_doms()<br/>\n";

    $error = [0, ''];
    foreach $query_id0 (keys %{$query->{'pairs'}}) { 
	foreach $query_id1 (keys %{$query->{'pairs'}->{$query_id0}}) { 
	    foreach $pair (@{$query->{'pairs'}->{$query_id0}->{$query_id1}}) { 
		($pair->{'error'} != 0) and next;

		# output a dom file
		$fn = join('', $tmp_dn, $pair->{'id'}, '.dom');
		if(!open($fh, ">$fn")) {
		    $pair->{'error'}++;
		    $error->[0] = 2;
		    $error->[1] .= "Cannot open '$fn' file for writing.<br/\n";
		    next;
		}
		$pair->{'dom_fn'} = $fn;
		print $fh $pair->{'doms'}->[0], "\n", $pair->{'doms'}->[1], "\n";
		close($fh);
	    }
	}
    }

    return $error;
}

sub n_contacts {
    my($cgi, $query) = @_;

    my $error;
    my $query_id0;
    my $query_id1;
    my $pair;
    my $cmd;
    my $fh;
    my $res1;
    my $res2;
    my @F;

    #print "# n_contacts()<br/>\n";

    $error = [0, ''];
    foreach $query_id0 (keys %{$query->{'pairs'}}) {
	foreach $query_id1 (keys %{$query->{'pairs'}->{$query_id0}}) {
	    foreach $pair (@{$query->{'pairs'}->{$query_id0}->{$query_id1}}) {
		($pair->{'error'} != 0) and next;

		# check whether or not the two parts of the template are in contact
		$cmd = "i2 -count -d " . $pair->{'dom_fn'};
		if(!open($fh, "$cmd |")) {
		    $error = [0, "cannot open pipe from '$cmd'."];
		    return $error;
		}
		$res1 = {};
		$res2 = {};
		while(<$fh>) {
		    (/^#/ or /\A\s*\Z/) and next;

		    # the bracketed stuff messes up splitting on whitespace, so remove it
		    s/\(.*?\)//g;
		    @F = split;

		    $res1->{$F[4]}++;
		    $res2->{$F[11]}++;
		}
		close($fh);

		# get the minimum number of residues in contact from either component
		($pair->{'contacts'}) = sort(scalar(keys(%{$res1})), scalar(keys(%{$res2})));
	    }
	}
    }

    return $error;
}

sub doms2seqs {
    my($cgi, $query) = @_;

    my $query_id0;
    my $query_id1;
    my $pair;
    my $dom_fn;
    my $seq_fn;
    my $cmd;
    my $in;
    my $seq;
    my $n_seqs;
    my $error;
    my $fh;
    my @pdbseq_errs;

    #print "# doms2seqs()<br/>\n";

    $error = [0, ''];
    foreach $query_id0 (keys %{$query->{'pairs'}}) { 
	foreach $query_id1 (keys %{$query->{'pairs'}->{$query_id0}}) {
	    foreach $pair (@{$query->{'pairs'}->{$query_id0}->{$query_id1}}) {
		($pair->{'error'} != 0) and next;
		($pair->{'contacts'} < 10) and next;

		$dom_fn = $pair->{'dom_fn'};
		$seq_fn = $dom_fn . '.aa';
		$cmd = "pdbseq -f $dom_fn -min 0 | perl -pe 's/^(>\\S+).*/\$1/;' > $seq_fn";

		if(Web::mysystem($cmd) != 0) {
		    $pair->{'error'}++;
		    $error->[0] = 2;
		    $error->[1] .= "'$cmd' failed: '$!'<br/>\n";
		    next;
		}

		# pdbseq returns 0 even when there are errors. The errors are reported on
		# STDOUT. Bio::SeqIO incorrectly generates two seq objects from this (but
		# fair enough, since it is expecting fasta format). So:
		@pdbseq_errs = ();
		if(!open($fh, $seq_fn)) {
		    $pair->{'error'}++;
		    $error->[0] = 2;
		    $error->[1] .= "Could not open '$seq_fn' file for reading.<br/>\n";
		    next;
		}
		while(<$fh>) {
		    /^error:\s*(.*)/ and push(@pdbseq_errs, sprintf("pdbseq failed for %s: '%s'.<br/>\n", $dom_fn, $1));
		}
		close($fh);
		if(@pdbseq_errs > 0) {
		    $pair->{'error'}++;
		    $error->[0] = 2;
		    $error->[1] = join('', $error->[1], @pdbseq_errs);
		    next;
		}

		$in = Bio::SeqIO->new('-file' => $seq_fn, '-format' => 'fasta');
		while($seq = $in->next_seq()) {
		    push @{$pair->{'domseqs'}}, $seq;
		}
		$n_seqs = scalar @{$pair->{'domseqs'}};
		if($n_seqs != 2) {
		    $pair->{'error'}++;
		    $error->[0] = 2;
		    $error->[1] .= sprintf("%d sequence%s read from '$seq_fn', but should be 2.<br/>\n", $n_seqs, ($n_seqs == 1) ? '' : 's');
		    next;
		}
	    }
	}
    }

    return $error;
}

sub align2templates {
    my($cgi, $query) = @_;

    my $error;
    my $query_id0;
    my $query_id1;
    my $pair;
    my $i;
    my $seq_fn;
    my $seq_out;
    my $aln_fn;
    my $cmd;
    my $aln_in;
    my $aln;
    my $subseq;
    my $seqs;

    #print "# align2templates()<br/>\n";

    $error = [0, ''];
    foreach $query_id0 (keys %{$query->{'pairs'}}) { 
	foreach $query_id1 (keys %{$query->{'pairs'}->{$query_id0}}) { 
	    $seqs = [$query->{'seqs'}->{$query_id0}, $query->{'seqs'}->{$query_id1}];
	    PAIR: foreach $pair (@{$query->{'pairs'}->{$query_id0}->{$query_id1}}) { 
		($pair->{'error'} != 0) and next;
		($pair->{'contacts'} < 10) and next;

		for($i = 0; $i < 2; $i++) {
		    # write the query sequence and the pdb domain sequence to a file

		    # FIXME - only output the hsp of the query
		    $subseq = Bio::Seq->new(
					    '-id'  => $seqs->[$i]->id,
					    '-seq' => $seqs->[$i]->subseq(
									  $pair->{'hsps'}->[$i]->start('query'),
									  $pair->{'hsps'}->[$i]->end('query'),
									 ),
					   );

		    $seq_fn = join('', $tmp_dn, $i, '.', $pair->{'id'}, '.qt.aa');
		    $seq_out = Bio::SeqIO->new('-file' => ">$seq_fn", '-format' => 'fasta');
		    $seq_out->write_seq($pair->{'domseqs'}->[$i]);
		    $seq_out->write_seq($subseq);

		    # align them
		    $aln_fn = join('', $tmp_dn, $i, '.', $pair->{'id'}, '.qt.muscle');

		    #$cmd = "muscle -quiet -stable -clwstrict -in $seq_fn -out $aln_fn";

		    $cmd = "muscle -quiet -clwstrict -in $seq_fn -out $aln_fn";# -stable not supported in muscle 3.8

		    if(Web::mysystem($cmd) != 0) {
			$pair->{'error'}++;
			$error->[0] = 2;
			$error->[1] .= sprintf(
					       "query pair %s-%s, muscle failed for %s/%d-%d on %s%s/%d-%d, cmd '$cmd': '%s'<br/>\n",
					       $query_id0,
					       $query_id1,
					       $seqs->[$i]->id,
					       $pair->{'hsps'}->[$i]->start('query'),
					       $pair->{'hsps'}->[$i]->end('query'),
					       $pair->{'pdb'}->{'id'},
					       $pair->{'chain_ids'}->[$i],
					       @{$pair->{'res'}->[$i]},
					       $!,
					      );
			next PAIR;
		    }

		    $aln_in = Bio::AlignIO->new('-file' => $aln_fn, '-format' => 'clustalw');
		    $aln = $aln_in->next_aln();
		    push @{$pair->{'aln_fns'}}, $aln_fn;
		    push @{$pair->{'alns'}}, $aln;
		}
	    }
	}
    }

    return $error;
}

sub run_interprets {
    my($cgi, $query) = @_;

    my $query_id0;
    my $query_id1;
    my $pair;
    my $fn;
    my $cmd;
    my $fh;
    my @F;
    my $id0;
    my $id1;
    my $raw;
    my $mean;
    my $sd;
    my $z_score;
    my $error;
    my $stat;
    my $ras_fn;
    my $err_fn;
    my $pcid0;
    my $pcid1;

    #print "# run_interprets()<br/>\n";

    $error = [0, ''];
    Q0: foreach $query_id0 (keys %{$query->{'pairs'}}) {
	Q1: foreach $query_id1 (keys %{$query->{'pairs'}->{$query_id0}}) {
	    PAIR: foreach $pair (@{$query->{'pairs'}->{$query_id0}->{$query_id1}}) {
		($pair->{'error'} != 0) and next;
		($pair->{'contacts'} < 10) and next;

		$fn = join('', $tmp_dn, $pair->{'id'}, '.i2.txt');
		$ras_fn = join('', $tmp_dn, $pair->{'id'}, '.i2.ras');
		$err_fn = join('', $tmp_dn, $pair->{'id'}, '.i2.err');
		$cmd = sprintf "i2 -mode 4 -rand 200 -a %s %s -d %s -ras %s 1> %s 2> $err_fn", @{$pair->{'aln_fns'}}, $pair->{'dom_fn'}, $ras_fn, $fn;
		$pair->{'i2'}->{'fn'} = $fn;
		$pair->{'i2'}->{'ras_fn'} = $ras_fn;

		# FIXME - use correct i2 error codes
		$stat = Web::mysystem($cmd);
		$stat >>= 8;
		if($stat != 0) {
		    $pair->{'error'}++;
		    $error->[0] = 2;

		    if(-s $err_fn) {
			if(open($fh, $err_fn)) {
			    while(<$fh>) {
				chomp;
				$error->[1] .= "$_<br/>\n";
			    }
			    close($fh);
			}
			else {
			    $error->[1] .= "Cannot open '$fn' file for reading.<br/>\n";
			    $error->[1] .= "'$cmd' failed with status $stat: '$!'<br/>\n";
			}
		    }
		    else {
			$error->[1] .= "'$cmd' failed with status $stat: '$!'<br/>\n";
		    }

		    #last Q0;
		    next;
		}
		unlink($err_fn);

		if(!open($fh, $fn)) {
		    $pair->{'error'}++;
		    $error->[0] = 2;
		    $error->[1] .= "Cannot open '$fn' file for reading.<br/>\n";
		    next;
		}

		$pair->{'i2'}->{'fn'} = $fn;
		while(<$fh>) {
#		    if(/FINAL_SCORE:\s+(\S+)\s+(\S+)\s+(\S+)\s+rand\s+(\S+)\s+(\S+)\s+(\S+)/) {
#			($id0, $id1, $raw, $mean, $sd, $z_score) = ($1, $2, $3, $4, $5, $6);
		    if(/SUM\*{0,1}\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+rand\s+(\S+)\s+(\S+)\s+(\S+)/) {
			($id0, $pcid0, $id1, $pcid1, $raw, $mean, $sd, $z_score) = ($1, $2, $3, $4, $5, $6, $7, $8);
			if(($id0 eq $query_id0) and ($id1 eq $query_id1)) {
			    $pair->{'i2'}->{'raw'}     = $raw;
			    $pair->{'i2'}->{'mean'}    = $mean;
			    $pair->{'i2'}->{'sd'}      = $sd;
			    $pair->{'i2'}->{'z_score'} = $z_score;
			    last;
			}
		    }
		}
		close($fh);
	    }
	}
    }

    return $error;
}

sub write_jmol {
    my($cgi, $query) = @_;

    my $query_id0;
    my $query_id1;
    my $pair;
    my $ras_fn;
    my $ras_fh;
    my $jmol_fn;
    my $jmol_fh;
    my $error;
    my $pdb_fn;
    my $pdb_url;
    my $script;
    my $aln_url0;
    my $aln_url1;
    my $i2_url;

    #print "# write_jmol()<br/>\n";

    $error = [0, ''];
    foreach $query_id0 (keys %{$query->{'pairs'}}) { 
	foreach $query_id1 (keys %{$query->{'pairs'}->{$query_id0}}) { 
	    foreach $pair (@{$query->{'pairs'}->{$query_id0}->{$query_id1}}) { 
		($pair->{'error'} != 0) and next;
		($pair->{'contacts'} < 10) and next;

		if($pair->{'i2'}->{'ras_fn'}) {
		    $ras_fn =  $pair->{'i2'}->{'ras_fn'};
		    if(!open($ras_fh, $ras_fn)) {
			$pair->{'error'}++;
			$error->[0] = 2;
			$error->[1] .= "Cannot open '$ras_fn' file for reading.<br/>\n";
			next;
		    }

		    $script = '';
		    while(<$ras_fh>) {
			chomp;
			if(/load (\S+)/) {
			    $pdb_fn = $1;
			    if(-e "/usr/local/apache/htdocs/pdb/$pdb_fn") {
				$pdb_url = "/pdb/$pdb_fn";
			    }
			    elsif(-e "/usr/local/apache/htdocs/pdb_obsolete/$pdb_fn") {
				$pdb_url = "/pdb_obsolete/$pdb_fn";
			    }
			    $_ = "load $pdb_url; spacefill off;"; # spacefill is on by default with jmol but not rasmol
			}
			$script .= "$_; ";
		    }
		    close($ras_fh);

		    $jmol_fn = join('', $tmp_dn, $pair->{'id'}, '.jmol.html');
		    if(!open($jmol_fh, ">$jmol_fn")) {
			$pair->{'error'}++;
			$error->[0] = 2;
			$error->[1] .= "Cannot open '$jmol_fn' file for writing.<br/>\n";
			next;
		    }

		    ($aln_url0 = $pair->{'aln_fns'}->[0]) =~ s/.*\///;
		    $aln_url0 = $tmp_dn_url . $aln_url0;

		    ($aln_url1 = $pair->{'aln_fns'}->[1]) =~ s/.*\///;
		    $aln_url1 = $tmp_dn_url . $aln_url1;

		    ($i2_url = $pair->{'i2'}->{'fn'}) =~ s/.*\///;
		    $i2_url = $tmp_dn_url . $i2_url;

		    Web::head($cgi, "InterPreTS $query_id0 with $query_id1" , 'interprets', $jmol_fh);

		    print $jmol_fh "<div id=\"title\">\n<h2>InterPreTS</h2>\n<p>\n<h4>$query_id0 with $query_id1</h4>\n";

		    printf(
			   $jmol_fh
			   "Z-score: <a href=\"%s\">%.3f</a> (raw = %.3f, &mu; = %.3f, &sigma; = %.3f)<br/>\n",
			   $i2_url,
			   $pair->{'i2'}->{'z_score'},
			   $pair->{'i2'}->{'raw'},
			   $pair->{'i2'}->{'mean'},
			   $pair->{'i2'}->{'sd'},
			  );

		    printf(
			   $jmol_fh
			   "Template: <a href=\"http://www.rcsb.org/pdb/explore/explore.do?structureId=%s\">%s</a> %s<br/>\n",
			   $pair->{'pdb'}->{'id'},
			   $pair->{'pdb'}->{'id'},
			   $pair->{'pdb'}->{'title'},
			  );

		    printf(
			   $jmol_fh
			   "Alignment 1: <a href=\"%s\">Query1/%d-%d on %s/%d-%d</a> (BLAST E = %.1e, %.1f%% id)<br/>\n",
			   $aln_url1,
			   $pair->{'hsps'}->[0]->start('query'),
			   $pair->{'hsps'}->[0]->end('query'),
			   $pair->{'chain_ids'}->[0],
			   @{$pair->{'res'}->[0]},
			   $pair->{'hsps'}->[0]->evalue,
			   $pair->{'hsps'}->[0]->frac_identical('total') * 100,
			  );

		    printf(
			   $jmol_fh
			   "Alignment 2: <a href=\"%s\">Query2/%d-%d on %s/%d-%d</a> (BLAST E = %.1e, %.1f%% id)<br/>\n",
			   $aln_url0,
			   $pair->{'hsps'}->[1]->start('query'),
			   $pair->{'hsps'}->[1]->end('query'),
			   $pair->{'chain_ids'}->[1],
			   @{$pair->{'res'}->[1]},
			   $pair->{'hsps'}->[1]->evalue,
			   $pair->{'hsps'}->[1]->frac_identical('total') * 100,
			  );

		    print(
			  $jmol_fh
			  "</p>\n",
			  "</div>",
			  "<div id=\"jmol\" class=\"jmol\">\n",
			  "<script>jmolInitialize(\"/jmol\");</script>\n",
			  "<script>jmolApplet(500, \"$script;\", \"i2\");</script>\n",
			  "</div>\n",
			 );
		    Web::tail($cgi, 'interprets', $jmol_fh);
		    close($jmol_fh);

		    $pair->{'i2'}->{'jmol_fn'} = $jmol_fn;
		}
	    }
	}
    }

    return $error;
}
